# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .datagrid import datagrid
from .PRM_gmtsar import PRM_gmtsar

class PRM(datagrid, PRM_gmtsar):

    # my replacement function for GMT based robust 2D trend coefficient calculations:
    # gmt trend2d r.xyz -Fxyzmw -N1r -V
    # gmt trend2d r.xyz -Fxyzmw -N2r -V
    # gmt trend2d r.xyz -Fxyzmw -N3r -V
    # https://github.com/GenericMappingTools/gmt/blob/master/src/trend2d.c#L719-L744
    # 3 model parameters
    # rank = 3 => nu = size-3
    @staticmethod
    def robust_trend2d(data, rank):
        """
        Perform robust linear regression to estimate the trend in 2D data.

        Parameters
        ----------
        data : numpy.ndarray
            Array containing the input data. The shape of the array should be (N, 3), where N is the number of data points.
            The first column represents the x-coordinates, the second column represents the y-coordinates (if rank is 3),
            and the third column represents the z-values.
        rank : int
            Number of model parameters to fit. Should be 1, 2, or 3. If rank is 1, the function fits a constant trend.
            If rank is 2, it fits a linear trend. If rank is 3, it fits a planar trend.

        Returns
        -------
        numpy.ndarray
            Array containing the estimated trend coefficients. The length of the array depends on the specified rank.
            For rank 1, the array will contain a single value (intercept).
            For rank 2, the array will contain two values (intercept and slope).
            For rank 3, the array will contain three values (intercept, slope_x, slope_y).

        Raises
        ------
        Exception
            If the specified rank is not 1, 2, or 3.

        Notes
        -----
        The function performs robust linear regression using the M-estimator technique. It iteratively fits a linear model
        and updates the weights based on the residuals until convergence. The weights are adjusted using Tukey's bisquare
        weights to downweight outliers.

        References
        ----------
        - Rousseeuw, P. J. (1984). Least median of squares regression. Journal of the American statistical Association, 79(388), 871-880.

        - Huber, P. J. (1973). Robust regression: asymptotics, conjectures and Monte Carlo. The Annals of Statistics, 1(5), 799-821.
        """
        import numpy as np
        from sklearn.linear_model import LinearRegression
        # scale factor for normally distributed data is 1.4826
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.median_abs_deviation.html
        MAD_NORMALIZE = 1.4826
        # significance value
        sig_threshold = 0.51

        if rank not in [1,2,3]:
            raise Exception('Number of model parameters "rank" should be 1, 2, or 3')

        #see gmt_stat.c
        def gmtstat_f_q (chisq1, nu1, chisq2, nu2):
            import scipy.special as sc

            if chisq1 == 0.0:
                return 1
            if chisq2 == 0.0:
                return 0
            return sc.betainc(0.5*nu2, 0.5*nu1, chisq2/(chisq2+chisq1))

        if rank in [2,3]:
            x = data[:,0]
            x = np.interp(x, (x.min(), x.max()), (-1, +1))
        if rank == 3:
            y = data[:,1]
            y = np.interp(y, (y.min(), y.max()), (-1, +1))
        z = data[:,2]
        w = np.ones(z.shape)

        if rank == 1:
            xy = np.expand_dims(np.zeros(z.shape),1)
        elif rank == 2:
            xy = np.expand_dims(x,1)
        elif rank == 3:
            xy = np.stack([x,y]).transpose()

        # create linear regression object
        mlr = LinearRegression()

        chisqs = []
        coeffs = []
        while True:
            # fit linear regression
            mlr.fit(xy, z, sample_weight=w)

            r = np.abs(z - mlr.predict(xy))
            chisq = np.sum((r**2*w))/(z.size-3)    
            chisqs.append(chisq)
            k = 1.5 * MAD_NORMALIZE * np.median(r)
            w = np.where(r <= k, 1, (2*k/r) - (k * k/(r**2)))
            sig = 1 if len(chisqs)==1 else gmtstat_f_q(chisqs[-1], z.size-3, chisqs[-2], z.size-3)
            # Go back to previous model only if previous chisq < current chisq
            if len(chisqs)==1 or chisqs[-2] > chisqs[-1]:
                coeffs = [mlr.intercept_, *mlr.coef_]

            #print ('chisq', chisq, 'significant', sig)
            if sig < sig_threshold:
                break

        # get the slope and intercept of the line best fit
        return (coeffs[:rank])

#     # standalone function is compatible but it is too slow while it should not be a problem
#     @staticmethod
#     def robust_trend2d(data, rank):
#         import numpy as np
#         import statsmodels.api as sm
# 
#         if rank not in [1, 2, 3]:
#             raise Exception('Number of model parameters "rank" should be 1, 2, or 3')
# 
#         if rank in [2, 3]:
#             x = data[:, 0]
#             x = np.interp(x, (x.min(), x.max()), (-1, +1))
#         if rank == 3:
#             y = data[:, 1]
#             y = np.interp(y, (y.min(), y.max()), (-1, +1))
#         z = data[:, 2]
# 
#         if rank == 1:
#             X = sm.add_constant(np.ones(z.shape))
#         elif rank == 2:
#             X = sm.add_constant(x)
#         elif rank == 3:
#             X = sm.add_constant(np.column_stack((x, y)))
# 
#         # Hampel weighting function looks the most accurate for noisy data
#         rlm_model = sm.RLM(z, X, M=sm.robust.norms.Hampel())
#         rlm_results = rlm_model.fit()
# 
#         return rlm_results.params[:rank]

#     # calculate MSE (optional)
#     @staticmethod
#     def robust_trend2d_mse(data, coeffs, rank):
#         import numpy as np
# 
#         x = data[:, 0]
#         y = data[:, 1]
#         z_actual = data[:, 2]
# 
#         if rank == 1:
#             z_predicted = coeffs[0]
#         elif rank == 2:
#             z_predicted = coeffs[0] + coeffs[1] * x
#         elif rank == 3:
#             z_predicted = coeffs[0] + coeffs[1] * x + coeffs[2] * y
# 
#         mse = np.mean((z_actual - z_predicted) ** 2)
#         return mse

    @staticmethod
    def from_list(prm_list):
        """
        Convert a list of parameter and value pairs to a PRM object.

        Parameters
        ----------
        prm_list : list
            A list of PRM strings.

        Returns
        -------
        PRM
            A PRM object.
        """
        from io import StringIO
        prm = StringIO('\n'.join(prm_list))
        return PRM._from_io(prm)

    @staticmethod
    def from_str(prm_string):
        """
        Convert a string of parameter and value pairs to a PRM object.

        Parameters
        ----------
        prm_string : str
            A PRM string.

        Returns
        -------
        PRM
            A PRM object.
        """
        from io import StringIO
        if isinstance(prm_string, bytes):
            # for cases like
            #return PRM.from_str(os.read(pipe2[0],int(10e6))
            prm_string = prm_string.decode('utf-8')
        # for cases like
        # return PRM.from_str(os.read(pipe2[0],int(10e6).decode('utf8'))
        prm = StringIO(prm_string)
        return PRM._from_io(prm)

    @staticmethod
    def from_file(prm_filename):
        """
        Convert a PRM file of parameter and value pairs to a PRM object.

        Parameters
        ----------
        prm_filename : str
            The filename of the PRM file.

        Returns
        -------
        PRM
            A PRM object.
        """
        #data = json.loads(document)
        prm = PRM._from_io(prm_filename)
        prm.filename = prm_filename
        return prm

    @staticmethod
    def _from_io(prm):
        """
        Read parameter and value pairs from IO stream to a PRM object.

        Parameters
        ----------
        prm : IO stream
            The IO stream.

        Returns
        -------
        PRM
            A PRM object.
        """
        import pandas as pd

        def to_numeric_or_original(vals):
            try:
                return pd.to_numeric(vals)
            except ValueError:
                return vals

        df = pd.read_csv(prm, sep='\s+=\s+', header=None, names=['name', 'value'], engine='python').set_index('name')
        df['value'] = df['value'].map(to_numeric_or_original)

        return PRM(df)

    def __init__(self, prm=None):
        """
        Initialize a PRM object.

        Parameters
        ----------
        prm : PRM or pd.DataFrame, optional
            The PRM object or DataFrame to initialize from. Default is None.

        Returns
        -------
        None
        """
        import pandas as pd

        def to_numeric_or_original(vals):
            try:
                return pd.to_numeric(vals)
            except ValueError:
                return vals

        # Initialize an empty DataFrame if prm is None
        if prm is None:
            _prm = pd.DataFrame(columns=['name', 'value'])
        elif isinstance(prm, pd.DataFrame):
            _prm = prm.reset_index()
        else:
            _prm = prm.df.reset_index()

        # Convert values to numeric where possible, keep original value otherwise
        _prm['value'] = _prm['value'].map(to_numeric_or_original)

        # Set the DataFrame for the PRM object
        self.df = _prm[['name', 'value']].drop_duplicates(keep='last').set_index('name')
        self.filename = None

    def __eq__(self, other):
        """
        Compare two PRM objects for equality.

        Parameters
        ----------
        other : PRM
            The other PRM object to compare with.

        Returns
        -------
        bool
            True if the PRM objects are equal, False otherwise.
        """
        return isinstance(self, PRM) and self.df == other.df

    def __str__(self):
        """
        Return a string representation of the PRM object.

        Returns
        -------
        str
            The string representation of the PRM object.
        """
        return self.to_str()

    def __repr__(self):
        """
        Return a string representation of the PRM object for debugging.

        Returns
        -------
        str
            The string representation of the PRM object. If the PRM object was created from a file, 
            the filename and the number of items in the DataFrame representation of the PRM object 
            are included in the string.
        """
        if self.filename:
            return 'Object %s (%s) %d items\n%r' % (self.__class__.__name__, self.filename, len(self.df), self.df)
        else:
            return 'Object %s %d items\n%r' % (self.__class__.__name__, len(self.df), self.df)

    # use 'g' format for Python and numpy float values
    def set(self, prm=None, gformat=False, **kwargs):
        """
        Set PRM values.

        Parameters
        ----------
        prm : PRM, optional
            The PRM object to set values from. Default is None.
        gformat : bool, optional
            Whether to use 'g' format for float values. Default is False.
        **kwargs
            Additional keyword arguments for setting individual values.

        Returns
        -------
        PRM
            The updated PRM object.
        """
        import numpy as np

        if isinstance(prm, PRM):
            for (key, value) in prm.df.itertuples():
                self.df.loc[key] = value
        elif prm is not None:
            raise Exception('Arguments is not a PRM object')
        for key, value in kwargs.items():
            self.df.loc[key] = float(format(value, 'g')) if gformat and type(value) \
                in [float, np.float16, np.float32, np.float64] else value
        return self

    def to_dataframe(self):
        """
        Convert the PRM object to a DataFrame.

        Returns
        -------
        pd.DataFrame
            The DataFrame representation of the PRM object.
        """
        return self.df

    def to_file(self, prm):
        """
        Save the PRM object to a PRM file.

        Parameters
        ----------
        prm : str
            The filename of the PRM file to save to.

        Returns
        -------
        PRM
            The PRM object.
        """
        self._to_io(prm)
        # update internal filename after saving with the new filename
        self.filename = prm
        return self

    #def update(self):
    #    if self.filename is None:
    #        raise Exception('PRM is not created from file, use to_file() method instead')
    #    return self._to_io(self.filename)
    def update(self, name=None, safe=False, debug=False):
        """
        Save PRM file to disk and rename all 3 linked files together: input, LED, SLC if "name" defined.
        If safe=True, save old (small) PRM and LED files and move only (large) SLC file.

        Parameters
        ----------
        name : str, optional
            The new name for the PRM file. Default is None.
        safe : bool, optional
            Whether to use safe mode. Default is False.
        debug : bool, optional
            Whether to enable debug mode. Default is False.

        Returns
        -------
        PRM
            The updated PRM object.
        """
        import shutil
        import os

        if self.filename is None:
            raise Exception('PRM is not created from file, use to_file() method instead')

        if debug:
            print ('Debug mode: only print expected operations but do not perform the actual job')

        # rename linked files
        if name is not None:
            # define current files directory
            dirname0 = os.path.dirname(self.filename)
            # define new files basename
            basename = os.path.splitext(name)[0]
            shortname = os.path.split(basename)[-1]

            # rename PRM file
            if not safe:
                print (f'Remove old PRM file {self.filename} and save new one {name}')
                if not debug:
                    os.remove(self.filename)
            #else:
            #    print (f'Safe mode: remain old PRM file {self.filename} and save new one {name}')
            # will be saved later
            self.filename = name

            input_file0 = os.path.join(dirname0, os.path.split(self.get('input_file'))[-1])
            input_ext = os.path.splitext(input_file0)[1]
            input_file = basename + input_ext
            #print (f'PRM change input_file attribute {input_file0} -> {input_file}')
            self.set(input_file = f'{shortname}{input_ext}')
            if os.path.isfile(input_file0) and not input_file == input_file0:
                if not safe:
                    #print (f'Rename input_file {input_file0} -> {input_file}')
                    if not debug:
                        os.rename(input_file0, input_file)
                else:
                    #print (f'Safe mode: copy input_file {input_file0} -> {input_file}')
                    if not debug:
                        shutil.copy2(input_file0, input_file, follow_symlinks=True)

            SLC_file0 = os.path.join(dirname0, os.path.split(self.get('SLC_file'))[-1])
            SLC_file = basename + '.SLC'
            #print (f'PRM change SLC_file attribute {SLC_file0} -> {SLC_file}')
            self.set(SLC_file = f'{shortname}.SLC')
            if os.path.isfile(SLC_file0) and not SLC_file == SLC_file0:
                #print (f'Rename SLC_file {SLC_file0} -> {SLC_file}')
                if not debug:
                    os.rename(SLC_file0, SLC_file)

            led_file0 = os.path.join(dirname0, os.path.split(self.get('led_file'))[-1])
            led_file = basename + '.LED'
            #print (f'PRM change LED_file attribute {led_file0} -> {led_file}')
            self.set(led_file = f'{shortname}.LED')
            if os.path.isfile(led_file0) and not led_file == led_file0:
                if not safe:
                    #print (f'Rename LED_file {led_file0} -> {led_file}')
                    if not debug:
                        os.rename(led_file0, led_file)
                else:
                    #print (f'Safe mode: copy LED_file {led_file0} -> {led_file}')
                    if not debug:
                        shutil.copy2(led_file0, led_file)

        if debug:
            return self
            #.sel('input_file','SLC_file','led_file')

        return self.to_file(self.filename)

    def to_str(self):
        """
        Convert the PRM object to a string.

        Returns
        -------
        str
            The PRM string.
        """
        return self._to_io()

    def _to_io(self, output=None):
        """
        Convert the PRM object to an IO stream.

        Parameters
        ----------
        output : IO stream, optional
            The IO stream to write the PRM string to. Default is None.

        Returns
        -------
        str
            The PRM string.
        """
        return self.df.reset_index().astype(str).apply(lambda row: (' = ').join(row), axis=1)\
            .to_csv(output, header=None, index=None)

    def sel(self, *args):
        """
        Select specific PRM attributes and create a new PRM object.

        Parameters
        ----------
        *args : str
            The attribute names to select.

        Returns
        -------
        PRM
            The new PRM object with selected attributes.
        """
        return PRM(self.df.loc[[*args]])

    def __add__(self, other):
        """
        Add two PRM objects or a PRM object and a scalar.

        Parameters
        ----------
        other : PRM or scalar
            The PRM object or scalar to add.

        Returns
        -------
        PRM
            The resulting PRM object after addition.
        """
        import pandas as pd
        if isinstance(other, PRM):
            prm = pd.concat([self.df, other.df])
            # drop duplicates
            prm = prm.groupby(prm.index).last()
        else:
            prm = self.df + other
        return PRM(prm)

    def __sub__(self, other):
        """
        Subtract two PRM objects or a PRM object and a scalar.

        Parameters
        ----------
        other : PRM or scalar
            The PRM object or scalar to subtract.

        Returns
        -------
        PRM
            The resulting PRM object after subtraction.
        """
        import pandas as pd
        if isinstance(other, PRM):
            prm = pd.concat([self.df, other.df])
            # drop duplicates
            prm = prm.groupby(prm.index).last()
        else:
            prm = self.df - other
        return PRM(prm)

    def get(self, *args):
        """
        Get the values of specific PRM attributes.

        Parameters
        ----------
        *args : str
            The attribute names to get values for.

        Returns
        -------
        Union[Any, List[Any]]
            The values of the specified attributes. If only one attribute is requested, 
            return its value directly. If multiple attributes are requested, return a list of values.
        """
        out = [self.df.loc[[key]].iloc[0].values[0] for key in args]
        if len(out) == 1:
            return out[0]
        return out

    def shift_atime(self, lines, inplace=False):
        """
        Shift time in azimuth by a number of lines.

        Parameters
        ----------
        lines : float
            The number of lines to shift by.
        inplace : bool, optional
            Whether to modify the PRM object in-place. Default is False.

        Returns
        -------
        PRM
            The shifted PRM object or DataFrame. If 'inplace' is True, returns modified PRM object,
            otherwise, returns a new PRM with shifted times.
        """
        prm = self.sel('clock_start','clock_stop','SC_clock_start','SC_clock_stop') + lines/self.get('PRF')/86400.0
        if inplace:
            return self.set(prm)
        else:
            return prm

    # fitoffset.csh 3 3 freq_xcorr.dat
    # PRM.fitoffset(3, 3, offset_dat)
    # PRM.fitoffset(3, 3, matrix_fromfile='raw/offset.dat')
    @staticmethod
    def fitoffset(rank_rng, rank_azi, matrix=None, matrix_fromfile=None, SNR=20):
        """
        Estimates range and azimuth offsets for InSAR (Interferometric Synthetic Aperture Radar) data.

        Parameters
        ----------
        rank_rng : int
            Number of parameters to fit in the range direction.
        rank_azi : int
            Number of parameters to fit in the azimuth direction.
        matrix : numpy.ndarray, optional
            Array of range and azimuth offset estimates. Default is None.
        matrix_fromfile : str, optional
            Path to a file containing range and azimuth offset estimates. Default is None.
        SNR : int, optional
            Signal-to-noise ratio cutoff. Data points with SNR below this threshold are discarded.
            Default is 20.

        Returns
        -------
        prm : PRM object
            An instance of the PRM class with the calculated parameters.

        Raises
        ------
        Exception
            If both 'matrix' and 'matrix_fromfile' arguments are provided or if neither is provided.
        Exception
            If there are not enough data points to estimate the parameters.

        Usage
        -----
        The function estimates range and azimuth offsets for InSAR data based on the provided input.
        It performs robust fitting to obtain range and azimuth coefficients, calculates scale coefficients,
        and determines the range and azimuth shifts. The resulting parameters are then stored in a PRM object.

        Example
        -------
        fitoffset(3, 3, matrix_fromfile='raw/offset.dat')
        """
        import numpy as np
        import math

        if (matrix is None and matrix_fromfile is None) or (matrix is not None and matrix_fromfile is not None):
            raise Exception('One and only one argument matrix or matrix_fromfile should be defined')
        if matrix_fromfile is not None:
            matrix = np.genfromtxt(matrix_fromfile)

        #  first extract the range and azimuth data
        rng = matrix[np.where(matrix[:,4]>SNR)][:,[0,2,1]]
        azi = matrix[np.where(matrix[:,4]>SNR)][:,[0,2,3]]

        # make sure there are enough points remaining
        if rng.shape[0] < 8:
            raise Exception(f'FAILED - not enough points to estimate parameters, try lower SNR ({rng.shape[0]} < 8)')

        rng_coef = PRM.robust_trend2d(rng, rank_rng)
        azi_coef = PRM.robust_trend2d(azi, rank_azi)

        # print MSE (optional)
        #rng_mse = PRM.robust_trend2d_mse(rng, rng_coef, rank_rng)
        #azi_mse = PRM.robust_trend2d_mse(azi, azi_coef, rank_azi)
        #print ('rng_mse_norm', rng_mse/len(rng), 'azi_mse_norm', azi_mse/len(azi))

        # range and azimuth data ranges
        scale_coef = [np.min(rng[:,0]), np.max(rng[:,0]), np.min(rng[:,1]), np.max(rng[:,1])]

        #print ('rng_coef', rng_coef)
        #print ('azi_coef', azi_coef)

        # now convert to range coefficients
        rshift = rng_coef[0] - rng_coef[1]*(scale_coef[1]+scale_coef[0])/(scale_coef[1]-scale_coef[0]) \
            - rng_coef[2]*(scale_coef[3]+scale_coef[2])/(scale_coef[3]-scale_coef[2])
        # now convert to azimuth coefficients
        ashift = azi_coef[0] - azi_coef[1]*(scale_coef[1]+scale_coef[0])/(scale_coef[1]-scale_coef[0]) \
            - azi_coef[2]*(scale_coef[3]+scale_coef[2])/(scale_coef[3]-scale_coef[2])
        #print ('rshift', rshift, 'ashift', ashift)

        # note: Python x % y expression and nympy results are different to C, use math function
        # use 'g' format for float values as in original GMTSAR codes to easy compare results
        prm = PRM().set(gformat=True,
                        rshift     =int(rshift) if rshift>=0 else int(rshift)-1,
                        sub_int_r  =math.fmod(rshift, 1)  if rshift>=0 else math.fmod(rshift, 1) + 1,
                        stretch_r  =rng_coef[1]*2/(scale_coef[1]-scale_coef[0]),
                        a_stretch_r=rng_coef[2]*2/(scale_coef[3]-scale_coef[2]),
                        ashift     =int(ashift) if ashift>=0 else int(ashift)-1,
                        sub_int_a  =math.fmod(ashift, 1)  if ashift>=0 else math.fmod(ashift, 1) + 1,
                        stretch_a  =azi_coef[1]*2/(scale_coef[1]-scale_coef[0]),
                        a_stretch_a=azi_coef[2]*2/(scale_coef[3]-scale_coef[2]),
                       )

        return prm

    def diff(self, other, gformat=True):
        """
        Compare the PRM object with another PRM object and return the differences.

        Parameters
        ----------
        other : PRM
            The other PRM object to compare with.
        gformat : bool, optional
            Whether to use 'g' format for float values. Default is True.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the differences between the two PRM objects.
        """
        import pandas as pd
        import numpy as np

        if not isinstance(other, PRM):
            raise Exception('Argument should be PRM class instance')

        df1 = self.df.copy()
        df2 = other.df.copy()

        if gformat:
            fmt = lambda v: format(v, 'g') if type(v) in [float, np.float16, np.float32, np.float64] else v
            df1['value'] = [fmt(value) for value in df1['value']]
            df2['value'] = [fmt(value) for value in df2['value']]

        return pd.concat([df1, df2]).drop_duplicates(keep=False)

    def fix_aligned(self):
        """
        Correction for the range and azimuth shifts of the re-aligned SLC images (fix_prm_params() in GMTSAR)
        """
        from scipy import constants
        # constant from GMTSAR code
        #SOL = 299792456.0
    
        delr = constants.speed_of_light / self.get('rng_samp_rate') / 2
        #delr = SOL / self.get('rng_samp_rate') / 2
        near_range = self.get('near_range') + \
            (self.get('st_rng_bin') - self.get('chirp_ext') + self.get('rshift') + self.get('sub_int_r') - 1)* delr
    
        SC_clock_start = self.get('SC_clock_start') + \
            (self.get('ashift') + self.get('sub_int_a')) / (self.get('PRF') * 86400.0) + \
            (self.get('nrows') - self.get('num_valid_az')) / (2 * self.get('PRF') * 86400.0)
    
        SC_clock_stop = SC_clock_start + \
            (self.get('num_valid_az') * self.get('num_patches')) / (self.get('PRF') * 86400.0)
    
        return self.set(near_range=near_range, SC_clock_start=SC_clock_start, SC_clock_stop=SC_clock_stop)

#     # note: only one dimension chunked due to sequential file writing 
#     def write_SLC_int(self, data):
#         import numpy as np
#         import os
# 
#         dirname = os.path.dirname(self.filename)
#         slc_filename = os.path.join(dirname, self.get('SLC_file'))
#         if os.path.exists(slc_filename):
#             os.remove(slc_filename)
# 
#         ys_blocks = np.array_split(np.arange(data.y.size), np.arange(0, data.y.size, self.chunksize)[1:])
#         with open(slc_filename, 'wb') as f:
#             for ys_block in ys_blocks:
#                 slc_block = data.isel(y=ys_block)
#                 re = slc_block.re.values
#                 im = slc_block.im.values
#                 buffer = np.zeros((2*re.size), dtype=re.dtype)
#                 buffer[::2] = re.ravel()
#                 buffer[1::2] = im.ravel()
#                 buffer.tofile(f)
#                 del buffer, re, im, slc_block

    # note: only one dimension chunked due to sequential file reading 
    def read_SLC_int(self, scale=2.5e-07):
        """
        Read SLC (Single Look Complex) data and compute the power of the signal.
        The method reads binary SLC data file, which contains alternating sequences of real and imaginary parts.
        It calculates the intensity of the signal and return it as a 2D numpy array.

        Returns
        -------
        xarray.DataArray
            2D array representing the power of the signal. The array shape corresponds to the dimensions of the SLC data.

        Notes
        -----
        This function uses a data factor (DFACT = 2.5e-07) from the GMTSAR code.
        The GMTSAR note indicates that the square of the intensity is used to match gips ihconv.
        The returned intensity data is flipped up-down ("shift data up if necessary") following the GMTSAR convention.

        Raises
        ------
        FileNotFoundError
            If the SLC file cannot be found.

        Example
        -------
        >>> import numpy as np
        >>> prm = PRM.from_file(filename)
        >>> amp = prm.read_SLC_int()
        """
        import xarray as xr
        import numpy as np
        import dask, dask.array
        import os
        import warnings

        @dask.delayed
        def read_SLC_block(slc_filename, start, stop):
            # Read a chunk of the SLC file
            # [real_0, imag_0, real_1, imag_1, real_2, imag_2, ...]
            #print ('    offset, shape', start*2, 2*(stop-start))
            # offset is measured in bytes
            return np.memmap(slc_filename, dtype=np.int16, mode='r', offset=start*4, shape=(2*(stop-start),))

        prm = PRM.from_file(self.filename)
        # num_patches multiplier is omitted
        slc_filename, xdim, ydim = prm.get('SLC_file', 'num_rng_bins', 'num_valid_az')
        dirname = os.path.dirname(self.filename)
        slc_filename = os.path.join(dirname, slc_filename)
        #print (slc_filename, ydim, xdim)

        blocksize = self.chunksize*xdim
        blocks = int(np.ceil(ydim * xdim / blocksize))
        #print ('chunks', chunks, 'chunksize', chunksize)
        # Create a lazy Dask array that reads chunks of the SLC file
        res = []
        ims = []
        for i in range(blocks):
            start = i * blocksize
            stop = min((i+1) * blocksize, ydim * xdim)
            #print ('start, stop, shape', start, stop, (stop-start))
            # use proper output data type for complex data and intensity
            block = dask.array.from_delayed(read_SLC_block(slc_filename, start, stop),
                shape=(2*(stop-start),), dtype=np.int16)
            res.append(block[::2])
            ims.append(block[1::2])
            del block
        # Concatenate the chunks together
        # Data file can include additional data outside of the specified dimensions
        re = dask.array.concatenate(res).reshape((-1, xdim))[:ydim,:]
        im = dask.array.concatenate(ims).reshape((-1, xdim))[:ydim,:]
        del res, ims
        coords = {'y': np.arange(ydim) + 0.5, 'x': np.arange(xdim) + 0.5}
        re = xr.DataArray(re, coords=coords).rename('re')
        im = xr.DataArray(im, coords=coords).rename('im')
        #if intensity:
        #    return ((scale or 1)*re.astype(np.float32))**2 + ((scale or 1)*im.astype(np.float32))**2
        if scale is not None:
            return scale * (xr.merge([re, im]).astype(np.float32))
        return xr.merge([re, im])

#     @staticmethod
#     def goldstein_filter(data, corr, psize):
#         import xarray as xr
#         import numpy as np
#         import dask
#     
#         def apply_pspec(data, alpha):
#             # NaN is allowed value
#             assert not(alpha < 0), f'Invalid parameter value {alpha} < 0'
#             wgt = np.power(np.abs(data)**2, alpha / 2)
#             data = wgt * data
#             return data
# 
#         def make_wgt(nxp, nyp):
#             # Create arrays of horizontal and vertical weights
#             wx = 1.0 - np.abs(np.arange(nxp // 2) - (nxp / 2.0 - 1.0)) / (nxp / 2.0 - 1.0)
#             wy = 1.0 - np.abs(np.arange(nyp // 2) - (nyp / 2.0 - 1.0)) / (nyp / 2.0 - 1.0)
#             # Compute the outer product of wx and wy to create the top-left quadrant of the weight matrix
#             quadrant = np.outer(wy, wx)
#             # Create a full weight matrix by mirroring the quadrant along both axes
#             wgt = np.block([[quadrant, np.flip(quadrant, axis=1)],
#                             [np.flip(quadrant, axis=0), np.flip(np.flip(quadrant, axis=0), axis=1)]])
#             return wgt
# 
#         def patch_goldstein_filter(data, corr, wgt, psize):
#             """
#             Apply the Goldstein adaptive filter to the given data.
# 
#             Args:
#                 data: 2D numpy array of complex values representing the data to be filtered.
#                 corr: 2D numpy array of correlation values. Must have the same shape as `data`.
# 
#             Returns:
#                 2D numpy array of filtered data.
#             """
#             # Calculate alpha
#             alpha = 1 - (wgt * corr).sum() / wgt.sum()
#             data = np.fft.fft2(data, s=(psize,psize))
#             data = apply_pspec(data, alpha)
#             data = np.fft.ifft2(data, s=(psize,psize))
#             return wgt * data
# 
#         def apply_goldstein_filter(data, corr, psize):
#             # Create an empty array for the output
#             out = np.zeros(data.shape, dtype=np.complex64)
#             # ignore processing for empty chunks 
#             if np.all(np.isnan(data)):
#                 return out
#             # Create the weight matrix
#             wgt_matrix = make_wgt(psize, psize)
#             # Iterate over windows of the data
#             for i in range(0, data.shape[0] - psize, psize // 2):
#                 for j in range(0, data.shape[1] - psize, psize // 2):
#                     # Create proocessing windows
#                     data_window = data[i:i+psize, j:j+psize]
#                     corr_window = corr[i:i+psize, j:j+psize]
#                     wgt_window = wgt_matrix[:data_window.shape[0],:data_window.shape[1]]
#                     # Apply the filter to the window
#                     filtered_window = patch_goldstein_filter(data_window, corr_window, wgt_window, psize)
#                     # Add the result to the output array
#                     slice_i = slice(i, min(i + psize, out.shape[0]))
#                     slice_j = slice(j, min(j + psize, out.shape[1]))
#                     out[slice_i, slice_j] += filtered_window[:slice_i.stop - slice_i.start, :slice_j.stop - slice_j.start]
#             return out
# 
#         # Apply function with overlap; psize//2 overlap is not enough (some empty lines produced)
#         # use complex data and real correlation
#         phase_filtered = dask.array.map_overlap(lambda data, corr: apply_goldstein_filter(data, corr, psize),
#                                        data.data,
#                                        corr.data,
#                                        depth=psize//2 + 2,
#                                        dtype=np.float32, 
#                                        meta=np.array(()))
#         # Calculate the phase and fix chunksizes
#         return xr.DataArray(np.arctan2(phase_filtered.imag, phase_filtered.real),coords=corr.coords)\
#                 .chunk(corr.chunksizes).rename('phase')

#     @staticmethod
#     def correlation(A1, A2, amp):
#         import xarray as xr
#         import numpy as np
#         # constant from GMTSAR code
#         thresh = 5.e-21
#         a = A1 * A2
#         corr = xr.where(a > 0, amp / np.sqrt(a), 0)
#         corr = xr.where(corr < 0, 0, corr)
#         corr = xr.where(corr > 1, 1, corr)
#         # mask too low amplitude areas as invalid
#         # amp1 and amp2 chunks are high for SLC, amp has normal chunks for NetCDF
#         return xr.where(a >= thresh, corr, np.nan).chunk(a.chunksizes).rename('phase')

#     # see about correlation filter
#     # https://github.com/gmtsar/gmtsar/issues/86
#     def intf(self, other, basedir, topo_fromfile, basename=None, \
#             wavelength=200, psi=False, psize=None, \
#             coarsen=(1,4), func=None, weight=None, \
#             phase_tofile='phase', corr_tofile='corr', debug=False):
#         """
#         Perform interferometric processing on the input SAR data.
# 
#         Parameters
#         ----------
#         other : PRM
#             Another instance of the PRM class representing the second SAR data.
#         basedir : str
#             Base directory for storing the processed data files.
#         topo_fromfile : str
#             Path to the topo file used for the calculation.
#         basename : str, optional
#             Base name for the output files. If not provided, a default name will be generated.
#         wavelength : int, optional
#             Wavelength of the SAR data in meters. Default is 200 m.
#         psi : bool, optional
#             PSI flag calls not apply wavelength to phase. Default is False.
#         psize : int, optional
#             Werner/Goldstein filter window size in pixels. Default is 32 pixels.
#         func : callable, optional
#             Custom function to apply on the processed data arrays. Default is None.
#         debug : bool, optional
#             Enable debug mode. Default is False.
# 
#         Returns
#         -------
#         None
# 
#         Notes
#         -----
#         This method performs interferometric processing on the input SAR data using GMTSAR tools.
#         It generates various intermediate and final files, including amplitude and phase data,
#         and applies filtering and convolution operations to produce the desired output.
#         """
#         import os
#         import numpy as np
#         import xarray as xr
#         import dask.array
#         import warnings
#         # suppress Dask warning "RuntimeWarning: divide by zero encountered in divide"
#         warnings.filterwarnings("ignore", category=RuntimeWarning, module="dask.core")
# 
#         if not isinstance(other, PRM):
#             raise Exception('Argument "other" should be PRM class instance')
# 
#         if psi:
#             #print ('NOTE: disable downscaling (coarsen=None) and Goldstein adaptive filtering (psize=None) in PSI mode')
#             #if wavelength is not None:
#             #    print ('NOTE: anti-aliasing (speckle) filtering applied to correlation only in PSI mode')
#             coarsen = None
#             psize = None
#         if psize is not None:
#             #print ('NOTE: apply downscaling (coarsen=(1,4)) for Goldstein adaptive filtering (psize)')
#             coarsen = (1,4)
# 
#         # expand simplified definition
#         if coarsen == 1 or coarsen == (1,1):
#             raise Exception('Argument "coarsen" should be None to disable downscaling')
#         if coarsen is not None and not isinstance(coarsen, (list,tuple, np.ndarray)):
#             coarsen = (coarsen, coarsen)
# 
#         # define basename from two PRM names
#         subswath = os.path.basename(self.filename)[16:18]
#         if basename is None:
#             # S1_20171111_ALL_F3.PRM -> F3
#             date1    = os.path.basename(self.filename)[3:11]
#             date2    = os.path.basename(other.filename)[3:11]
#             basename = f'{subswath}_{date1}_{date2}_'
#             #print ('basename', basename)
# 
#         # make full file name, use workaround for 'weight' argument name defined without extension
#         fullname = lambda name: os.path.join(basedir, basename + name if name[-4:]=='.grd' else f'{subswath}_{name}.grd')
# 
#         # prepare PRMs for the calculation below
#         other.set(self.SAT_baseline(other, tail=9))
#         self.set(self.SAT_baseline(self).sel('SC_height','SC_height_start','SC_height_end'))
# 
#         # for topo use relative path from PRM files directory
#         # use imag.grd=bf for GMT native, C-binary format
#         # os.path.join(basedir, f'{subswath}_topo.grd')
#         if os.path.exists(fullname('real.grd')):
#             os.remove(fullname('real.grd'))
#         if os.path.exists(fullname('imag.grd')):
#             os.remove(fullname('imag.grd'))
#         self.phasediff(other, topo_fromfile=topo_fromfile,
#                        imag_tofile=fullname('imag.grd'),
#                        real_tofile=fullname('real.grd'),
#                        debug=debug)
# 
#         # original SLC (do not flip vertically)
#         amp1 = self.read_SLC_int(intensity=True)
#         amp2 = other.read_SLC_int(intensity=True)
#         # phasediff tool output files (flip vertically)
#         imag = xr.open_dataarray(fullname('imag.grd'), engine=self.engine, chunks=self.chunksize)
#         imag.data = dask.array.flipud(imag)
#         real = xr.open_dataarray(fullname('real.grd'), engine=self.engine, chunks=self.chunksize)
#         real.data = dask.array.flipud(real)
#         #real.shape, imag.shape (5484, 21572) (5484, 21572)
#         #print ('DEBUG X1 real.shape, imag.shape', real.shape, imag.shape)
# 
#         # anti-aliasing filter for multi-looking, wavelength can be None
#         imag_filt = self.antialiasing_downscale(imag, weight=weight, wavelength=wavelength, coarsen=coarsen, debug=debug)
#         real_filt = self.antialiasing_downscale(real, weight=weight, wavelength=wavelength, coarsen=coarsen, debug=debug)
#         amp1_filt = self.antialiasing_downscale(amp1, weight=weight, wavelength=wavelength, coarsen=coarsen, debug=debug)
#         amp2_filt = self.antialiasing_downscale(amp2, weight=weight, wavelength=wavelength, coarsen=coarsen, debug=debug)
#         # cleanup
#         del amp1, amp2
#         
#         # calculate amplitude of interferogram
#         amp_filt = np.sqrt(real_filt**2 + imag_filt**2)
#         # calculate masked correlation
#         corr = self.correlation(amp1_filt, amp2_filt, amp_filt)
#         # cleanup
#         del amp1_filt, amp2_filt, amp_filt
#         #chunksize=(512, 5393)
#         #print ('DEBUG X2 corr', corr)
# 
#         # Apply Goldstein filter function after multi-looking
#         if debug:
#             print ('DEBUG: intf apply Goldstein filter with size', psize is not None, psize)
#         if psize is not None:
#             phase = self.goldstein_filter((real + 1j * imag), corr, psize=psize)
#         elif psi:
#             phase = np.arctan2(imag, real)
#         else:
#             phase = np.arctan2(imag_filt, real_filt)
#         # cleanup
#         del real, imag, real_filt, imag_filt
#         #chunksize=(512, 5393)
#         #print ('DEBUG X3 phase', phase)
# 
#         if func is not None:
#             corr = func(corr)
#         corr_filename = fullname(f'{corr_tofile}.grd')
#         if os.path.exists(corr_filename):
#             os.remove(corr_filename)
#         encoding = {'corr': self._compression(corr.shape)}
#         #print ('DEBUG X2', corr_da)
#         # rename to save lazy NetCDF preventing broken coordinates (y,y) 
#         corr.rename('corr').rename({'y': 'a', 'x': 'r'}).to_netcdf(corr_filename, encoding=encoding, engine=self.engine)
# 
#         if func is not None:
#             phase = func(phase)
#         phase_filename = fullname(f'{phase_tofile}.grd')
#         if os.path.exists(phase_filename):
#             os.remove(phase_filename)
#         encoding = {'phase': self._compression(phase.shape)}
#         #print ('DEBUG X3', phase_da)
#         # mask phase using masked correlation
#         # rename to save lazy NetCDF preventing broken coordinates (y,y)
#         phase.rename('phase').rename({'y': 'a', 'x': 'r'}).to_netcdf(phase_filename, encoding=encoding, engine=self.engine)
# 
#         # cleanup
#         del corr, phase
#         for name in ['real.grd', 'imag.grd']:
#             filename = fullname(name)
#             if os.path.exists(filename):
#                 os.remove(filename)
# 
#         #print ('DEBUG X4')
#         return

    # see make_gaussian_filter.c for the original code
    def get_spacing(self, grid=1):
        """
        Calculate azimuth and range pixel size in meters.

        Parameters
        ----------
        grid : tuple or xarray.DataArray, optional
            A pair of x, y grid decimation coefficients or a 2D or 3D Xarray DataArray representing the decimation grid.
            The default is 1 for the input Sentinel-1 SLC grid.

        Returns
        -------
        tuple
            A tuple containing the azimuth and range pixel sizes in meters.

        Notes
        -----
        This method calculates the pixel sizes in the azimuth and range directions based on the provided parameters.
        The azimuth pixel size is determined using the spacecraft velocity, height, and pulse repetition frequency (PRF),
        while the range pixel size is calculated based on the speed of light and the range sampling rate.
        """
        import xarray as xr
        import numpy as np
        from scipy.constants import speed_of_light

        # define grid spacing
        if isinstance(grid, (xr.DataArray, xr.Dataset)):
            assert self.is_ra(grid), 'Raster needs to be defined in radar coordinates'
            dy = grid.y.diff('y')[0].item()
            dx = grid.x.diff('x')[0].item()
        elif isinstance(grid, (list, tuple)):
            dy, dx = grid
        else:
            dy = dx = grid
            
        # compute the range and azimuth pixel size
        RE, vel, ht, prf, fs, near_range, num_rng_bins = \
            self.get('earth_radius','SC_vel', 'SC_height', 'PRF', 'rng_samp_rate', 'near_range', 'num_rng_bins')
        #ER, vel, ht, prf, fs, near_range, num_rng_bins
        # real_vel/prf
        azi_px_size = vel / np.sqrt(1 + ht / RE) / prf
        rng_px_size = speed_of_light / fs / 2.0
        # compute the cosine of the looking angle and the surface deviate angle
        a = ht + RE
        far_range = near_range + rng_px_size * num_rng_bins
        rng = (near_range + far_range) / 2.0
        cost = (np.power(a, 2.0) + np.power(rng, 2.0) - np.power(RE, 2.0)) / 2.0 / a / rng
        cosa = (np.power(a, 2.0) + np.power(RE, 2.0) - np.power(rng, 2.0)) / 2.0 / a / RE
        # compute the ground range pixel size
        rng_px_size = rng_px_size / np.sin(np.arccos(cost) + np.arccos(cosa))
        # ground spacing in meters
        return (dy * azi_px_size, dx * rng_px_size)

#     # TODO: use PRM parameters to define config parameters
#     def snaphu_config(self, defomax=0, **kwargs):
#         """
#         Generate a configuration file for Snaphu phase unwrapping.
# 
#         Parameters
#         ----------
#         defomax : int, optional
#             Maximum number of deformation cycles allowed. Default is 0.
#         **kwargs
#             Additional configuration parameters in key-value format.
# 
#         Returns
#         -------
#         str
#             The generated configuration file as a string.
# 
#         Examples
#         --------
#         Get default SNAPHU config:
#         config = sbas.snaphu_config()
# 
#         Get custom SNAPHU config with added tiling options:
#         config = sbas.snaphu_config(defomax=DEFOMAX, NTILEROW=1, NTILECOL=2, ROWOVRLP=200, COLOVRLP=200)
# 
#         Notes
#         -----
#         This method generates a configuration file for Snaphu, a phase unwrapping software.
#         The configuration file includes basic parameters and allows customization by passing additional parameters.
# 
#         Custom Parameters
#         -----------------
#         Additional configuration parameters can be passed as keyword arguments (e.g., `parameter_name=value`).
#         Boolean values should be provided as `True` or `False`.
#         """
#         import os
#         # we already use joblib everywhere
#         import joblib
# 
#         tiledir = os.path.splitext(self.filename)[0]
#         n_jobs = joblib.cpu_count()
# 
#         conf_basic = f"""
#         # basic config
#         INFILEFORMAT   FLOAT_DATA
#         OUTFILEFORMAT  FLOAT_DATA
#         AMPFILEFORMAT  FLOAT_DATA
#         CORRFILEFORMAT FLOAT_DATA
#         ALTITUDE       693000.0
#         EARTHRADIUS    6378000.0
#         NEARRANGE      831000
#         DR             18.4
#         DA             28.2
#         RANGERES       28
#         AZRES          44
#         LAMBDA         0.0554658
#         NLOOKSRANGE    1
#         NLOOKSAZ       1
#         TILEDIR        {tiledir}_snaphu_tiledir
#         NPROC          {n_jobs}
#         """
#         conf_custom = '# custom config\n'
#         # defomax can be None
#         keyvalues = ([('DEFOMAX_CYCLE', defomax)] if defomax is not None else []) + list(kwargs.items())
#         for key, value in keyvalues:
#             if isinstance(value, bool):
#                 value = 'TRUE' if value else 'FALSE'
#             conf_custom += f'        {key} {value}\n'
#         return conf_basic + conf_custom
# 



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .tqdm_joblib import tqdm_joblib

class XYZTiles(tqdm_joblib):

    def download(self, geometry, zoom, n_jobs=-1, url='https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}'):
        """
        from pygmtsar import XYZTiles
        gmap = XYZTiles().download(AOI.geometry[0], zoom = 17)
        gmap.plot.imshow()
        """
        import xarray as xr
        import requests
        import imageio.v3 as iio
        import math
        import io
        import numpy as np
        from tqdm.auto import tqdm
        import joblib

        def deg2num(lat_deg, lon_deg, zoom):
            lat_rad = math.radians(lat_deg)
            n = 2.0 ** zoom
            x = int((lon_deg + 180.0) / 360.0 * n)
            y = int((1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * n)
            return (x, y)

        def download_tile(url, x, y, z):
            #url = f'https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}'
            response = requests.get(url.format(x=x, y=y, z=z))
            image = iio.imread(io.BytesIO(response.content))
            return image

        def num2deg(xtile, ytile, zoom):
            n = 2.0 ** zoom
            lon_deg = xtile / n * 360.0 - 180.0
            lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
            lat_deg = math.degrees(lat_rad)
            return (lat_deg, lon_deg)

        # latitudes inverted
        #lat_start, lat_end = 34.42, 34.36
        #lon_start, lon_end = -118.5, -118.4
        if isinstance(geometry, (xr.DataArray, xr.Dataset)):
            lon_start = geometry.lon.min()
            lat_start = geometry.lat.min()
            lon_end   = geometry.lon.max()
            lat_end   = geometry.lat.max()
        else:
            lon_start, lat_start, lon_end, lat_end = geometry.bounds
        lat_start, lat_end = lat_end, lat_start

        # Calculate tile range
        x_start, y_start = deg2num(lat_start, lon_start, zoom)
        x_end, y_end = deg2num(lat_end, lon_end, zoom)

    #     def job_tile(x, y):
    #         tile_image = download_tile(url, x, y, zoom)
    #         tile_array = np.array(tile_image)
    #         lat_upper_left, lon_upper_left = num2deg(x, y, zoom)
    #         lat_lower_right, lon_lower_right = num2deg(x + 1, y + 1, zoom)
    #         latitudes = np.linspace(lat_upper_left, lat_lower_right, tile_array.shape[0])
    #         longitudes = np.linspace(lon_upper_left, lon_lower_right, tile_array.shape[1])
    #         return xr.DataArray(tile_array, dims=('lat', 'lon', 'band'), 
    #                                coords={'lat': latitudes, 'lon': longitudes}).rename('tile')
        def job_tile(x, y):
            tile_image = download_tile(url, x, y, zoom)
            tile_array = np.array(tile_image)
            lat_upper_left, lon_upper_left = num2deg(x, y, zoom)
            lat_lower_right, lon_lower_right = num2deg(x + 1, y + 1, zoom)
            # Exclude the endpoint to prevent overlap with the adjacent tile
            latitudes = np.linspace(lat_upper_left, lat_lower_right, tile_array.shape[0], endpoint=False)
            longitudes = np.linspace(lon_upper_left, lon_lower_right, tile_array.shape[1], endpoint=False)
            return xr.DataArray(tile_array, dims=('lat', 'lon', 'band'), 
                                coords={'lat': latitudes, 'lon': longitudes}).rename('colors')

        with self.tqdm_joblib(tqdm(desc='XYZ Tile Downloading', total=(x_end-x_start+1)*(y_end-y_start+1))) as progress_bar:
            tile_xarrays = joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(job_tile)(x, y)\
                                for x in range(x_start, x_end + 1) for y in range(y_start, y_end + 1))

        return xr.combine_by_coords(tile_xarrays)['colors'].transpose('band', 'lat', 'lon')



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2021, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_dem import Stack_dem
from .PRM import PRM

class Stack_align(Stack_dem):

    def _offset2shift(self, xyz, rmax, amax, method='linear'):
        """
        Convert offset coordinates to shift values on a grid.

        Parameters
        ----------
        xyz : numpy.ndarray
            Array containing the offset coordinates (x, y, z).
        rmax : int
            Maximum range bin.
        amax : int
            Maximum azimuth line.
        method : str, optional
            Interpolation method. Default is 'linear'.

        Returns
        -------
        xarray.DataArray
            Array containing the shift values on a grid.
        """
        import xarray as xr
        import numpy as np
        from scipy.interpolate import griddata

        # use center pixel GMT registration mode
        rngs = np.arange(8/2, rmax+8/2, 8)
        azis = np.arange(4/2, amax+4/2, 4)
        grid_r, grid_a = np.meshgrid(rngs, azis)

        grid = griddata((xyz[:,0], xyz[:,1]), xyz[:,2], (grid_r, grid_a), method=method)
        da = xr.DataArray(np.flipud(grid), coords={'y': azis, 'x': rngs}, name='z')
        return da

    # replacement for gmt grdfilter ../topo/dem.grd -D2 -Fg2 -I12s -Gflt.grd
    # use median decimation instead of average
    def _get_topo_llt(self, subswath, degrees, debug=False):
        """
        Get the topography coordinates (lon, lat, z) for decimated DEM.

        Parameters
        ----------
        subswath : int
            Subswath number.
        degrees : float
            Number of degrees for decimation.
        debug : bool, optional
            Enable debug mode. Default is False.

        Returns
        -------
        numpy.ndarray
            Array containing the topography coordinates (lon, lat, z).
        """
        import xarray as xr
        import numpy as np
        import warnings
        # supress warnings "UserWarning: The specified chunks separate the stored chunks along dimension"
        warnings.filterwarnings('ignore')

        # add buffer around the cropped area for borders interpolation
        dem_area = self.get_dem(subswath)
        ny = int(np.round(degrees/dem_area.lat.diff('lat')[0]))
        nx = int(np.round(degrees/dem_area.lon.diff('lon')[0]))
        if debug:
            print ('DEBUG: DEM decimation','ny', ny, 'nx', nx)
        dem_area = dem_area.coarsen({'lat': ny, 'lon': nx}, boundary='pad').mean()

        lats, lons, z = xr.broadcast(dem_area.lat, dem_area.lon, dem_area)
        topo_llt = np.column_stack([lons.values.ravel(), lats.values.ravel(), z.values.ravel()])
        return topo_llt

    # aligning for reference image
    def _align_ref_subswath(self, subswath, debug=False):
        """
        Align and stack the reference scene.

        Parameters
        ----------
        subswath : int
            Subswath number.
        debug : bool, optional
            Enable debug mode. Default is False.

        Returns
        -------
        None

        Examples
        --------
        stack.stack_ref(subswath=2, debug=True)
        """
        import xarray as xr
        import numpy as np
        import os

        reference_line = list(self.get_reference(subswath).itertuples())[0]
        #print (reference_line)

        # for reference scene
        multistem, stem = self.multistem_stem(subswath, reference_line.datetime)
        path_stem = os.path.join(self.basedir, stem)
        path_multistem = os.path.join(self.basedir, multistem)

        # generate PRM, LED, SLC
        self._make_s1a_tops(subswath, debug=debug)

        PRM.from_file(path_stem + '.PRM')\
            .set(input_file = path_multistem + '.raw')\
            .update(path_multistem + '.PRM', safe=True)

        self._ext_orb_s1a(subswath, multistem, debug=debug)

        # recalculate after _ext_orb_s1a
        earth_radius = PRM.from_file(path_multistem + '.PRM')\
            .calc_dop_orb(inplace=True).update().get('earth_radius')

    # aligning for secondary image
    def _align_rep_subswath(self, subswath, date=None, degrees=12.0/3600, debug=False):
        """
        Align and stack secondary images.

        Parameters
        ----------
        subswath : int
            Subswath number.
        date : str or None, optional
            Date of the image to process. If None, process all images. Default is None.
        degrees : float, optional
            Degrees per pixel resolution for the coarse DEM. Default is 12.0/3600.
        debug : bool, optional
            Enable debug mode. Default is False.

        Returns
        -------
        None

        Examples
        --------
        stack.stack_rep(subswath=2, date='2023-05-01', degrees=15.0/3600, debug=True)
        """
        import xarray as xr
        import numpy as np
        import os
        
        # temporary filenames to be removed
        cleanup = []

        reference_line = list(self.get_reference(subswath).itertuples())[0]
        multistem, stem = self.multistem_stem(subswath, reference_line.datetime)
        #print (reference_line)

        # define reference image parameters
        # TODO: use reference PRM filename instead of subswath
        reference = self.PRM(subswath=subswath).sel('earth_radius').set(stem=stem, multistem=multistem)

        # prepare coarse DEM for alignment
        # 12 arc seconds resolution is enough, for SRTM 90m decimation is 4x4
        topo_llt = self._get_topo_llt(subswath, degrees=degrees)
        #topo_llt.shape

        line = list(self.get_repeat(subswath, date).itertuples())[0]
        multistem, stem = self.multistem_stem(subswath, line.datetime)
        #print (line)

        # define relative filenames for PRM
        stem_prm    = os.path.join(self.basedir, stem + '.PRM')
        mstem_prm   = os.path.join(self.basedir, multistem + '.PRM')
        reference_prm  = os.path.join(self.basedir, reference.get("stem") + '.PRM')
        mreference_prm = os.path.join(self.basedir, reference.get("multistem") + '.PRM')

        # TODO: define 1st image for line, in the example we have no more
        tmp_da = 0

        # generate PRM, LED
        self._make_s1a_tops(subswath, date, debug=debug)

        # compute the time difference between first frame and the rest frames
        t1, prf = PRM.from_file(stem_prm).get('clock_start', 'PRF')
        t2      = PRM.from_file(stem_prm).get('clock_start')
        nl = int((t2 - t1)*prf*86400.0+0.2)
        #echo "Shifting the reference PRM by $nl lines..."

        # Shifting the reference PRM by $nl lines...
        # shift the super-references PRM based on $nl so SAT_llt2rat gives precise estimate
        prm1 = PRM.from_file(reference_prm)
        prm1.set(prm1.sel('clock_start' ,'clock_stop', 'SC_clock_start', 'SC_clock_stop') + nl/prf/86400.0)
        tmp_prm = prm1

        # compute whether there are any image offset
        #if tmp_da == 0:
        # tmp_prm defined above from {reference}.PRM
        prm1 = tmp_prm.calc_dop_orb(reference.get('earth_radius'), inplace=True, debug=debug)
        prm2 = PRM.from_file(stem_prm).calc_dop_orb(reference.get('earth_radius'), inplace=True, debug=debug).update()
        lontie,lattie = prm1.SAT_baseline(prm2, debug=debug).get('lon_tie_point', 'lat_tie_point')
        tmp_am = prm1.SAT_llt2rat(coords=[lontie, lattie, 0], precise=1, debug=debug)[1]
        tmp_as = prm2.SAT_llt2rat(coords=[lontie, lattie, 0], precise=1, debug=debug)[1]
        # bursts look equal to rounded result int(np.round(...))
        tmp_da = int(tmp_as - tmp_am)
        #print ('tmp_am', tmp_am, 'tmp_as', tmp_as, 'tmp_da', tmp_da)

        # in case the images are offset by more than a burst, shift the super-reference's PRM again
        # so SAT_llt2rat gives precise estimate
        if abs(tmp_da) >= 1000:
            prf = tmp_prm.get('PRF')
            tmp_prm.set(tmp_prm.sel('clock_start' ,'clock_stop', 'SC_clock_start', 'SC_clock_stop') - tmp_da/prf/86400.0)
            #raise Exception('TODO: Modifying reference PRM by $tmp_da lines...')

        # tmp.PRM defined above from {reference}.PRM
        prm1 = tmp_prm.calc_dop_orb(reference.get('earth_radius'), inplace=True, debug=debug)
        tmpm_dat = prm1.SAT_llt2rat(coords=topo_llt, precise=1, debug=debug)
        prm2 = PRM.from_file(stem_prm).calc_dop_orb(reference.get('earth_radius'), inplace=True, debug=debug)
        tmp1_dat = prm2.SAT_llt2rat(coords=topo_llt, precise=1, debug=debug)

        # get r, dr, a, da, SNR table to be used by fitoffset.csh
        offset_dat0 = np.hstack([tmpm_dat, tmp1_dat])
        func = lambda row: [row[0],row[5]-row[0],row[1],row[6]-row[1],100]
        offset_dat = np.apply_along_axis(func, 1, offset_dat0)

        # define radar coordinates extent
        rmax, amax = PRM.from_file(stem_prm).get('num_rng_bins','num_lines')

        # prepare the offset parameters for the stitched image
        # set the exact borders in radar coordinates
        par_tmp = offset_dat[(offset_dat[:,0]>0) & (offset_dat[:,0]<rmax) & (offset_dat[:,2]>0) & (offset_dat[:,2]<amax)]
        par_tmp[:,2] += nl
        if abs(tmp_da) >= 1000:
            par_tmp[:,2] -= tmp_da
            par_tmp[:,3] += tmp_da

        # prepare the rshift and ashift look up table to be used by make_s1a_tops
        # use tmp_dat instead of offset_dat
        r_xyz = offset_dat[:,[0,2,1]]
        a_xyz = offset_dat[:,[0,2,3]]

        r_grd = self._offset2shift(r_xyz, rmax, amax)
        r_grd_filename = stem_prm[:-4]+'_r.grd'
        r_grd.to_netcdf(r_grd_filename, engine=self.netcdf_engine)
        # drop the temporary file at the end of the function
        cleanup.append(r_grd_filename)

        a_grd = self._offset2shift(a_xyz, rmax, amax)
        a_grd_filename = stem_prm[:-4]+'_a.grd'
        a_grd.to_netcdf(a_grd_filename, engine=self.netcdf_engine)
        # drop the temporary file at the end of the function
        cleanup.append(a_grd_filename)

        # generate the image with point-by-point shifts
        # note: it removes calc_dop_orb parameters from PRM file
        # generate PRM, LED
        self._make_s1a_tops(subswath,
                           date=line.Index, mode=1,
                           rshift_fromfile=f'{stem}_r.grd',
                           ashift_fromfile=f'{stem}_a.grd',
                           debug=debug)

        # need to update shift parameter so stitch_tops will know how to stitch
        PRM.from_file(stem_prm).set(PRM.fitoffset(3, 3, offset_dat)).update()

        # echo stitch images together and get the precise orbit
        # use stitch_tops tmp.stitchlist $stem to merge images

        # the raw file does not exist but it works
        PRM.from_file(stem_prm)\
            .set(input_file = f'{multistem}.raw')\
            .update(mstem_prm, safe=True)

        self._ext_orb_s1a(subswath, multistem, date=line.Index, debug=debug)

        # Restoring $tmp_da lines shift to the image... 
        PRM.from_file(mstem_prm).set(ashift=0 if abs(tmp_da) < 1000 else tmp_da, rshift=0).update()

        # that is safe to rewrite source files
        prm1 = PRM.from_file(mreference_prm)
        prm1.resamp(PRM.from_file(mstem_prm),
                    repeatSLC_tofile=mstem_prm[:-4]+'.SLC',
                    interp=1, debug=debug
        ).to_file(mstem_prm)

        PRM.from_file(mstem_prm).set(PRM.fitoffset(3, 3, par_tmp)).update()
        # TEST
        #prm1.set(PRM.fitoffset(3, 3, par_tmp)).update()

        PRM.from_file(mstem_prm).calc_dop_orb(reference.get('earth_radius'), 0, inplace=True, debug=debug).update()
        # TEST
        #prm1.calc_dop_orb(reference.get('earth_radius'), 0, inplace=True, debug=debug).update()
        
        # cleanup
        for filename in cleanup:
            #if os.path.exists(filename):
            os.remove(filename)

#     # merge_swath.c modified for SLCs
#     def merge_date(self, date, chunksize=None, debug=False):
#         import xarray as xr
#         import numpy as np
#         import os
#         from scipy import constants
# 
#         if chunksize is None:
#             chunksize = chunksize
# 
#         subswaths = self.get_subswaths()
# 
#         # define offset parameters to merge subswaths
#         prms = []
#         for subswath in subswaths:
#             #print (subswath)
#             prm = self.PRM(date, subswath=subswath)
#             prms.append(prm)
# 
#         assert len(np.unique([prm.get('PRF') for prm in prms])), 'Image PRFs are not consistent'
#         assert len(np.unique([prm.get('rng_samp_rate') for prm in prms])), 'Image range sampling rates are not consistent'
# 
#         heads = [0] + [((prm.get('clock_start') - prms[0].get('clock_start')) * 86400 * prms[0].get('PRF')).round().astype(int) for prm in prms[1:]]
#         # head123: 0, 466, -408
#         if debug:
#             print ('heads', heads)
#         # minh: -408
#         minh = min(heads)
#         if debug:
#             print ('minh', minh)
# 
#         #head123: 408, 874, 0
#         heads = np.asarray(heads) - minh
#         if debug:
#             print ('heads', heads)
# 
#         #ovl12,23: 2690, 2558
#         ovls = [prm1.get('num_rng_bins') - ((prm2.get('near_range') - prm1.get('near_range')) / (constants.speed_of_light/ prm1.get('rng_samp_rate') / 2)).round().astype(int) for (prm1, prm2) in zip(prms[:-1], prms[1:])]
#         if debug:
#             print ('ovls', ovls)
# 
#         #Writing the grid files..Size(69158x13075)...
#         #maxy: 13075
#         # TODO: disable offset 1
#         maxy = max([prm.get('num_valid_az') + head for prm, head in zip(prms, heads)]) + 1
#         # for SLC
#         #maxy = max([prm.get('num_valid_az') + head for prm, head in zip(prms, heads)])
#         if debug:
#             print ('maxy', maxy)
#         maxx = sum([prm.get('num_rng_bins') - ovl -1 for prm, ovl in zip(prms, [-1] + ovls)])
#         if debug:
#             print ('maxx', maxx)
# 
#         # disable heads shift
#         #head123: 467, 1, 875
#         #heads = [maxy - prm.get('num_valid_az') - head for prm, head in zip(prms, heads)]
#         #print ('heads', heads)
# 
#         #Stitching location n1 = 1045
#         #Stitching location n2 = 935
#         ns = [np.ceil(-prm.get('rshift') + prm.get('first_sample') + 150.0).astype(int) for prm in prms[1:]]
#         ns = [10 if n < 10 else n for n in ns]
#         if debug:
#             print ('ns', ns)
#         #ns [1070, 963]
#         
#         # TODO: add subswath_offset and subswath_head to PRM files
# 
#         # TEST: disable 1st subswath head offset to check phasediff
#         #heads[0] = 400
# 
#         # merge
#         slcs = []
#         # left and right coordinates for every subswath valid area
#         x1s = []
#         x2s = []
#         
#         # 1st
#         xlim = prms[0].get('num_rng_bins') - ovls[0] + ns[0]
#         x1s.append(0)
#         x2s.append(xlim)
#         # disable scaling
#         slc = prms[0].read_SLC_int(scale=None)
#         slc = slc.isel(x=slice(None,xlim)).assign_coords(y=slc.y + heads[0])
#         slcs.append(slc)
# 
#         # 2nd
#         if len(prms) == 2:
#             # TODO: check for 2 subswaths
#             xlim = -1
#         else:
#             # for 3 subswaths
#             xlim = prms[1].get('num_rng_bins') - ovls[1] + ns[1]
#         x1s.append(ns[0])
#         x2s.append(xlim)
#         # disable scaling
#         slc = prms[1].read_SLC_int(scale=None)
#         slc = slc.isel(x=slice(ns[0],xlim)).assign_coords(y=slc.y + heads[1])
#         slcs.append(slc)
# 
#         # 3rd
#         if len(prms) == 3:
#             # disable scaling
#             slc = prms[2].read_SLC_int(scale=None)
#             x1s.append(ns[1])
#             x2s.append(-2)
#             slc = slc.isel(x=slice(ns[1],-2)).assign_coords(y=slc.y + heads[2])
#             slcs.append(slc)
# 
#         # check and merge SLCs
#         assert maxx == sum([slc.x.size for slc in slcs]), 'Incorrect output grid range dimension size'
#         slc = xr.concat(slcs, dim='x', fill_value=0).assign_coords(x=0.5 + np.arange(maxx))
#         assert slc.y.size == maxy, 'Incorrect output grid azimuth dimension size'
#         assert slc.x.size == maxx, 'Incorrect output grid range dimension sizes'
# 
#         # define merge filenames
#         subswath = ''.join(map(str, subswaths))
#         line = self.get_repeat(subswaths[0], date).iloc[0]
#         multistem, stem = self.multistem_stem(subswath, line.datetime)
#         #print (multistem, stem)
#         prm_filename = os.path.join(self.basedir, multistem + '.PRM')
#         if debug:
#             print ('prm_filename', prm_filename)
# 
#         # merge PRM
#         prm = PRM(prms[0])
#         dt = -minh / prm.get('PRF') / 86400
#         prm = prm.set(SLC_file=multistem + '.SLC',
#                       num_lines=maxy, nrows=maxy, num_valid_az=maxy,
#                       num_rng_bins=maxx, bytes_per_line=4*maxx, good_bytes=4*maxx,
#                       SC_clock_start=prm.get('SC_clock_start') - dt,
#                       clock_start=prm.get('clock_start') - dt,
#                       SC_clock_stop=prm.get('SC_clock_start') + maxy / prm.get('PRF') / 86400,
#                       clock_stop=prm.get('clock_start') + maxy / prm.get('PRF') / 86400)\
#                       .calc_dop_orb(prm.get('earth_radius'), 0, inplace=True, debug=debug)\
#                       .to_file(prm_filename)
# 
# #         # merge PRM - test updating parameters for the central area
# #         prm = PRM(prms[0])
# #         dt = -minh / prm.get('PRF') / 86400
# #         near_range      = np.mean([prm.get('near_range') for prm in prms])
# #         SC_vel          = np.mean([prm.get('SC_vel') for prm in prms])
# #         SC_height       = np.mean([prm.get('SC_height') for prm in prms])
# #         SC_height_start = prms[0].get('SC_height_start')
# #         SC_height_end   = prms[-1].get('SC_height_end')
# #         prm = prm.set(SLC_file=multistem + '.SLC',
# #                       num_lines=maxy, nrows=maxy, num_valid_az=maxy,
# #                       num_rng_bins=maxx, bytes_per_line=4*maxx, good_bytes=4*maxx,
# #                       SC_clock_start=prm.get('SC_clock_start') - dt,
# #                       clock_start=prm.get('clock_start') - dt,
# #                       SC_clock_stop=prm.get('SC_clock_start') + maxy / prm.get('PRF') / 86400,
# #                       clock_stop=prm.get('clock_start') + maxy / prm.get('PRF') / 86400,
# #                       near_range=near_range, SC_vel=SC_vel, SC_height=SC_height,
# #                       SC_height_start=SC_height_start, SC_height_end=SC_height_end).to_file(prm_filename)
#         #return prm, slc
#         # save merged SLC
#         prm.write_SLC_int(slc, chunksize=chunksize)
# 
#         # add calculated offsets to single subswaths
#         for idx, prm in enumerate(prms):
#             prm.set(smath_maxy=maxy, swath_maxx=maxx,
#                     swath_minh=minh, swath_head=heads[idx],
#                     swath_left=x1s[idx], swath_right=x2s[idx]).update()
# 
#         # cleanup
#         # [os.path.join(self.basedir, prm.get('led_file')) for prm in prms]
#         # [prm.filename for prm in prms]
#         cleanup = [os.path.join(self.basedir, prm.get('SLC_file')) for prm in prms]
#         for filename in cleanup:
#             if debug:
#                 print ('DEBUG: remove', filename)
#             os.remove(filename)

    # define bottoms from reference scene and apply for all scenes
    def get_subswaths_offsets(self, date, offsets=None, debug=False):
        import xarray as xr
        import numpy as np
        from scipy import constants

        subswaths = self.get_subswaths()

        # define offset parameters to merge subswaths
        prms = []
        for subswath in subswaths:
            #print (subswath)
            prm = self.PRM(date, subswath=subswath)
            prms.append(prm)

        assert len(np.unique([prm.get('PRF') for prm in prms])), 'Image PRFs are not consistent'
        assert len(np.unique([prm.get('rng_samp_rate') for prm in prms])), 'Image range sampling rates are not consistent'

        if offsets is None:
            bottoms = [0] + [((prm.get('clock_start') - prms[0].get('clock_start')) * 86400 * prms[0].get('PRF')).round().astype(int) for prm in prms[1:]]
            # head123: 0, 466, -408
            if debug:
                print ('bottoms init', bottoms)
            # minh: -408
            minh = min(bottoms)
            if debug:
                print ('minh', minh)
            #head123: 408, 874, 0
            bottoms = np.asarray(bottoms) - minh
        else:
            bottoms = offsets['bottoms']
            minh = offsets['bottom']
        if debug:
            print ('bottoms', bottoms)

        #ovl12,23: 2690, 2558
        ovls = [prm1.get('num_rng_bins') - ((prm2.get('near_range') - prm1.get('near_range')) \
                    / (constants.speed_of_light/ prm1.get('rng_samp_rate') / 2)).round().astype(int) \
                for (prm1, prm2) in zip(prms[:-1], prms[1:])]
        if debug:
            print ('ovls', ovls)

        #Writing the grid files..Size(69158x13075)...
        #maxy: 13075
        # for SLC
        maxy = max([prm.get('num_valid_az') + bottom for prm, bottom in zip(prms, bottoms)])
        if debug:
            print ('maxy', maxy)
        maxx = sum([prm.get('num_rng_bins') - ovl -1 for prm, ovl in zip(prms, [-1] + ovls)])
        if debug:
            print ('maxx', maxx)

        #Stitching location n1 = 1045
        #Stitching location n2 = 935
        ns = [np.ceil(-prm.get('rshift') + prm.get('first_sample') + 150.0).astype(int) for prm in prms[1:]]
        ns = [10 if n < 10 else n for n in ns]
        if debug:
            print ('ns', ns)
        #ns [1070, 963]

        # left and right coordinates for every subswath valid area
        x1s = []
        x2s = []

        # 1st
        xlim = prms[0].get('num_rng_bins') - ovls[0] + ns[0]
        x1s.append(0)
        x2s.append(xlim)

        # 2nd
        if len(prms) == 2:
            xlim = prms[1].get('num_rng_bins') - 1
        else:
            # for 3 subswaths
            xlim = prms[1].get('num_rng_bins') - ovls[1] + ns[1]
        x1s.append(ns[0])
        x2s.append(xlim)

        # 3rd
        if len(prms) == 3:
            xlim = prms[2].get('num_rng_bins') - 2
            x1s.append(ns[1])
            x2s.append(xlim)

        # check and merge SLCs
        sumx = sum([right-left for right, left in zip(x2s, x1s)])
        if debug:
            print ('assert maxx == sum(...)', maxx, sumx)
        assert maxx == sumx, 'Incorrect output grid range dimension size'

        # create reference merged PRM for geocoding to extent radar coordinates calculation
        filename = prms[0].filename[:-5] + ''.join(map(str, subswaths))
        prm_filename = filename + '.PRM'
        #print ('prm_filename', prm_filename)
        prm = PRM(prms[0])
        dt = -minh / prm.get('PRF') / 86400
        prm = prm.set(SLC_file=None,
                      num_lines=maxy, nrows=maxy, num_valid_az=maxy,
                      num_rng_bins=maxx, bytes_per_line=4*maxx, good_bytes=4*maxx,
                      SC_clock_start=prm.get('SC_clock_start') - dt,
                      clock_start=prm.get('clock_start') - dt,
                      SC_clock_stop=prm.get('SC_clock_start') + maxy / prm.get('PRF') / 86400,
                      clock_stop=prm.get('clock_start') + maxy / prm.get('PRF') / 86400)\
            .to_file(prm_filename)

        return {'bottoms': bottoms, 'lefts': x1s, 'rights': x2s, 'bottom': minh, 'extent': [maxy, maxx]}

    # merge_swath.c modified for SLCs
    # use reference scene vertical subswath aligments
    def _merge_subswaths(self, date, offsets, xlim1=None, ylim1=None, xlim2=None, ylim2=None, debug=False):
        import xarray as xr
        import numpy as np
        import os

        subswaths = self.get_subswaths()

        #if debug:
        #    print ('offsets', offsets)
        # use reference scene offsets
        maxy = offsets['extent'][0]
        minh = offsets['bottom']
        bottoms = offsets['bottoms']
        maxx = offsets['extent'][1]
        lefts = offsets['lefts']
        rights = offsets['rights']

        slcs = []
        prms = []
        for subswath, bottom, left, right in zip(subswaths, bottoms, lefts, rights):
            prm = self.PRM(date, subswath=subswath)
            # disable scaling
            slc = prm.read_SLC_int(scale=None)
            slc = slc.isel(x=slice(left, right)).assign_coords(y=slc.y + bottom)
            slcs.append(slc)
            prms.append(prm)

        # check and merge SLCs, use zero fill for np.int16 datatype
        slc = xr.concat(slcs, dim='x', fill_value=0).assign_coords(x=0.5 + np.arange(maxx))

        if debug:
            print ('assert slc.y.size == maxy', slc.y.size, maxy)
        assert slc.y.size == maxy, 'Incorrect output grid azimuth dimension size'
        if debug:
            print ('assert slc.x.size == maxx', slc.x.size, maxx)
        assert slc.x.size == maxx, 'Incorrect output grid range dimension sizes'
        del slcs

        # define merge filenames  
        filename = prms[0].filename[:-5] + ''.join(map(str, subswaths))
        prm_filename = filename + '.PRM'
        netcdf_filename = os.path.basename(filename + '.grd')
        if debug:
            print ('prm_filename', prm_filename, 'netcdf_filename', netcdf_filename)

        # merge PRM
        prm = PRM(prms[0])
        dt = -minh / prm.get('PRF') / 86400
        prm = prm.set(SLC_file=None, netcdf_filename=netcdf_filename,
                      num_lines=maxy, nrows=maxy, num_valid_az=maxy,
                      num_rng_bins=maxx, bytes_per_line=4*maxx, good_bytes=4*maxx,
                      SC_clock_start=prm.get('SC_clock_start') - dt,
                      clock_start=prm.get('clock_start') - dt,
                      SC_clock_stop=prm.get('SC_clock_start') + maxy / prm.get('PRF') / 86400,
                      clock_stop=prm.get('clock_start') + maxy / prm.get('PRF') / 86400)\
                 .to_file(prm_filename)
        #.calc_dop_orb(prm.get('earth_radius'), 0, inplace=True, debug=debug)\

        # add PRM to grid
        #slc.attrs['prm'] = str(prm)

        # save merged SLC
        #prm.write_SLC_int(slc, chunksize=chunksize)
        # encoding = {vn: self._compression(slc[vn].shape, complevel=0,
#                     chunksize=(self.chunksize, 4*self.chunksize)) for vn in slc.data_vars}
        encoding = {vn: self._compression(slc[vn].shape) for vn in slc.data_vars}
        # add directory name
        netcdf_filename = os.path.join(self.basedir, netcdf_filename)
        # rename dimensions to prevent issue with square output
        slc.rename({'y': 'a', 'x': 'r'})\
            .sel(a=slice(ylim1, ylim2), r=slice(xlim1, xlim2))\
            .to_netcdf(netcdf_filename, encoding=encoding, engine=self.netcdf_engine)

        # add calculated offsets to single subswaths
    #         for idx, prm in enumerate(prms):
    #             prm.set(smath_maxy=maxy, swath_maxx=maxx,
    #                     swath_bottom=bottoms[idx],
    #                     swath_left=lefts[idx], swath_right=rights[idx]).update()
        # for single SLCs add crop coordinates
        #slc.attrs['bottom'] = bottom[idx]
        #slc.attrs['left'] = left[idx]
        #slc.attrs['right'] = rights[idx]

        # cleanup
        # [os.path.join(self.basedir, prm.get('led_file')) for prm in prms]
        # [prm.filename for prm in prms]
        cleanup = [os.path.join(self.basedir, prm.get('SLC_file')) for prm in prms]
        for filename in cleanup:
            if debug:
                print ('DEBUG: remove', filename)
            os.remove(filename)

#     # merge_swath.c modified for SLCs
#     # use reference scene vertical subswath aligments
#     def aligh_subswaths(self, date, offsets, debug=False):
#         import numpy as np
#         import os
# 
#         subswaths = self.get_subswaths()
# 
#         #offsets = self.get_subswaths_offsets(date, offsets=offsets, debug=debug)
#         #if debug:
#         #    print ('offsets', offsets)
#         # use reference scene vertical offsets
#         maxy = offsets['extent'][0]
#         minh = offsets['bottom']
#         bottoms = offsets['bottoms']
#         maxx = offsets['extent'][1]
#         lefts = offsets['lefts']
#         rights = offsets['rights']
# 
#         prms = []
#         for subswath, bottom, left, right in zip(subswaths, bottoms, lefts, rights):
#             prm = self.PRM(date, subswath=subswath)
#             prms.append(prm)
# 
#         # define merge filenames  
#         prm_filename = prms[0].filename[:-5] + ''.join(map(str, subswaths)) + '.PRM'
#         if debug:
#             print ('prm_filename', prm_filename)
# 
#         # merge PRM
#         prm = PRM(prms[0])
#         dt = -minh / prm.get('PRF') / 86400
#         prm = prm.set(SLC_file=None,
#                       swath_bottom=';'.join(map(str, bottoms)),
#                       swath_left=';'.join(map(str, lefts)),
#                       swath_right=';'.join(map(str, rights)),
#                       num_lines=maxy, nrows=maxy, num_valid_az=maxy,
#                       num_rng_bins=maxx, bytes_per_line=4*maxx, good_bytes=4*maxx,
#                       SC_clock_start=prm.get('SC_clock_start') - dt,
#                       clock_start=prm.get('clock_start') - dt,
#                       SC_clock_stop=prm.get('SC_clock_start') + maxy / prm.get('PRF') / 86400,
#                       clock_stop=prm.get('clock_start') + maxy / prm.get('PRF') / 86400)\
#                  .to_file(prm_filename)
#         #.calc_dop_orb(prm.get('earth_radius'), 0, inplace=True, debug=debug)\

    def _convert_subswath(self, subswath, date, xlim1=None, ylim1=None, xlim2=None, ylim2=None, debug=False):
        import xarray as xr
        #import numpy as np
        import os

        prm = self.PRM(date, subswath=subswath)
        # crop the full grid if needed
        slc = prm.read_SLC_int(scale=None).sel(y=slice(ylim1, ylim2), x=slice(xlim1, xlim2))
        # add PRM to grid
        #slc.attrs['prm'] = str(prm)

        netcdf_filename = prm.filename[:-4] + '.grd'
        #print ('netcdf_filename', netcdf_filename)
        # cleanup before saving
        if os.path.exists(netcdf_filename):
            os.remove(netcdf_filename)
        # complevel=0 means disabled compression and the fastest saving
        encoding = {vn: self._compression(slc[vn].shape) for vn in slc.data_vars}
        # rename dimensions to prevent issue with square output
        slc.rename({'y': 'a', 'x': 'r'})\
            .to_netcdf(netcdf_filename, encoding=encoding, engine=self.netcdf_engine)

        # cleanup
        slc_filename = os.path.join(self.basedir, prm.get('SLC_file'))
        if debug:
            print ('DEBUG: remove', slc_filename)
        # file always exists, no check required
        os.remove(slc_filename)

    def baseline_table(self, n_jobs=-1, debug=False):
        """
        Generates a baseline table for Sentinel-1 data, containing dates and baseline components.

        This function creates a baseline table for Sentinel-1 data by processing the PRM files, which
        contain metadata for each image. The table includes dates and parallel and perpendicular
        baseline components for each image.

        Parameters
        ----------
        n_jobs : int, optional
            Number of CPU cores to use for parallel processing (default is -1, which means using all available cores).
        debug : bool, optional
            If True, print additional information during processing (default is False).

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the baseline table with date, times, and baseline components.

        Notes
        -----
        This function processes Sentinel-1 data by first generating PRM and LED files if they don't exist,
        then calculating doppler and orbital parameters for each image, and finally computing the baseline
        components. The routine is suited to be used before alignment to detect the best reference scene.

        """
        import pandas as pd
        import numpy as np
        from tqdm.auto import tqdm
        import joblib
        import os

        # use any one subswath in case of many
        subswath = self.get_subswaths()[0]
        datetimes = self.df[self.df.subswath==subswath].datetime

        def get_filename(dt):
            stem = self.multistem_stem(subswath, dt)[1]
            filename = os.path.join(self.basedir, f'{stem}.PRM')
            return filename

        def ondemand(date, dt):
            if not os.path.exists(get_filename(dt)):
                self._make_s1a_tops(subswath, date, debug=debug)

        # generate PRM, LED if needed
        #for (date, dt) in datetimes.iteritems():
        #    #print (dt, date)
        #    ondemand(dt)
        with self.tqdm_joblib(tqdm(desc='PRM generation', total=len(datetimes))) as progress_bar:
            joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(ondemand)(date, dt) for (date, dt) in datetimes.items())

        # calc_dop_orb() required for SAT_baseline
        reference_dt = datetimes[self.reference]
        prm_ref = PRM().from_file(get_filename(reference_dt)).calc_dop_orb(inplace=True)
        data = []
        for (date, dt) in datetimes.items():
            prm_rep = PRM().from_file(get_filename(dt))
            BPL, BPR = prm_ref.SAT_baseline(prm_rep).get('B_parallel', 'B_perpendicular')
            data.append({'date':date, 'parallel':BPL.round(1), 'perpendicular':BPR.round(1)})
        return pd.DataFrame(data).set_index('date')

    # 'threading' for Docker and 'loky' by default
    def compute_align(self, geometry='auto', dates=None, n_jobs=-1, degrees=12.0/3600, joblib_aligning_backend=None, debug=False):
        """
        Stack and align scenes.

        Parameters
        ----------
        dates : list or None, optional
            List of dates to process. If None, process all scenes. Default is None.
        n_jobs : int, optional
            Number of parallel processing jobs. n_jobs=-1 means all processor cores are used. Default is -1.
        joblib_aligning_backend: str or None, optional

        Returns
        -------
        None

        Examples
        --------
        stack.align()
        """
        import numpy as np
        import geopandas as gpd
        from tqdm.auto import tqdm
        import joblib
        import warnings
        # supress warnings about unary_union future behaviour to replace None by empty collection 
        warnings.filterwarnings('ignore')

        if dates is None:
            dates = self.df.index.unique()
        dates_rep = [date for date in dates if date != self.reference]

        subswaths = self.get_subswaths()

        # prepare reference scene
        #self.stack_ref()
        with self.tqdm_joblib(tqdm(desc='Aligning Reference', total=len(subswaths))) as progress_bar:
            joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(self._align_ref_subswath)(subswath, debug=debug) for subswath in subswaths)

        # prepare secondary images
        with self.tqdm_joblib(tqdm(desc='Aligning Repeat', total=len(dates_rep)*len(subswaths))) as progress_bar:
            # threading backend is the only one working inside Docker container to run multiple binaries in parallel
            joblib.Parallel(n_jobs=n_jobs, backend=joblib_aligning_backend)(joblib.delayed(self._align_rep_subswath)(subswath, date, degrees=degrees, debug=debug) \
                                           for date in dates_rep for subswath in subswaths)

        if len(subswaths) > 1:
            # calculate the offsets and also create merged reference PRM
            offsets = self.get_subswaths_offsets(self.reference, debug=debug)
            # DEM extent in radar coordinates, merged reference PRM required
            extent_ra = self.get_extent_ra()
            minx, miny, maxx, maxy = np.round(extent_ra.bounds).astype(int)
            #print ('minx, miny, maxx, maxy', minx, miny, maxx, maxy)
            with self.tqdm_joblib(tqdm(desc=f'Merging Subswaths', total=len(dates))) as progress_bar:
                joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(self._merge_subswaths)(date, offsets, minx, miny, maxx, maxy, debug=debug) \
                                               for date in dates)
        else:
            # DEM extent in radar coordinates, merged reference PRM required
            extent_ra = self.get_extent_ra()
            minx, miny, maxx, maxy = np.round(extent_ra.bounds).astype(int)
            #print ('minx, miny, maxx, maxy', minx, miny, maxx, maxy)
            # in case of a single subswath only convert SLC to NetCDF grid
            with self.tqdm_joblib(tqdm(desc='Convert Subswath', total=len(dates))) as progress_bar:
                joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(self._convert_subswath)(subswaths[0], date, minx, miny, maxx, maxy, debug=debug) \
                                               for date in dates)

        # merge subswaths, datapath and metapath converted to lists even for a single subswath, geometry merges bursts
        df = self.df.groupby(self.df.index).agg({'datetime': 'min', 'orbit': 'min', 'mission': 'min', 'polarization': 'min',
                                            'subswath': lambda s: int(''.join(map(str,list(s)))),
                                            'datapath': lambda p: list(p),
                                            'metapath': lambda p: list(p),
                                            'orbitpath': 'min',
                                            'geometry': lambda g: g.unary_union
                                           })
        # update the main object for the merged subswaths
        self.df = gpd.GeoDataFrame(df)
        


# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .tqdm_joblib import tqdm_joblib

class ASF(tqdm_joblib):
    from datetime import timedelta

    template_url = 'https://datapool.asf.alaska.edu/SLC/S{satellite}/{scene}.zip'
    template_safe = '*.SAFE/*/{mission}-iw{subswath}-slc-{polarization}-*'
    # URL of the HTML page
    # see https://asf.alaska.edu/data-sets/sar-data-sets/sentinel-1/sentinel-1-data-and-imagery/
    # https://s1qc.asf.alaska.edu
    poeorb_url = 'https://s1qc.asf.alaska.edu/aux_poeorb/'
    resorb_url = 'https://s1qc.asf.alaska.edu/aux_resorb/'
    # resorb - 69108
    # poeorb - 9811
    #pattern_orbit  = r'S1\w_OPER_AUX_\w{3}ORB_OPOD_\d{8}T\d{6}_V\d{8}T\d{6}_\d{8}T\d{6}.EOF'
    pattern_orbit  = r'S1\w_OPER_AUX_(POE|RES)ORB_OPOD_\d{8}T\d{6}_V\d{8}T\d{6}_\d{8}T\d{6}.EOF'
    template_orbit = '{mission}_OPER_AUX_(POE|RES)ORB_OPOD_\d+T\d+_V{date_start}T\d+_{date_end}T\d+.EOF'
    # see _select_orbit.py in sentineleof package
    #Orbital period of Sentinel-1 in seconds
    #T_ORBIT = (12 * 86400.0) / 175.0
    resorb_start_offset = timedelta(seconds=(12 * 86400.0) // 175.0 + 60)
    resorb_end_offset = timedelta(seconds=300)
    poeorb_start_offset = timedelta(seconds=0)
    poeorb_end_offset = timedelta(seconds=0)
    
    def __init__(self, username=None, password=None):
        import asf_search
        import getpass
        if username is None:
            username = getpass.getpass('Please enter your ASF username and press Enter key:')
        if password is None:
            password = getpass.getpass('Please enter your ASF password and press Enter key:')
        self.username = username
        self.password = password

    def _get_orbits(self, scenes, session, kind):
        import pandas as pd
        import re
        from datetime import datetime

        if kind == 'RESORB':
            url = self.resorb_url
            orbit_start_offset = self.resorb_start_offset
            orbit_end_offset = self.resorb_end_offset
        elif kind == 'POEORB':
            url = self.poeorb_url
            orbit_start_offset = self.poeorb_start_offset
            orbit_end_offset = self.poeorb_end_offset
        else:
            raise Exception(f'Invalid orbit type: {kind}.')
        #print ('url', url)
        # get orbits HTML list
        responce = session.get(url)
        if responce.status_code == 200:
            lines = responce.text.splitlines()
            # TODO: cleanup code
            orbits = [re.search(self.pattern_orbit, line).group(0) if re.search(self.pattern_orbit, line) else None for line in lines]
            orbits = pd.DataFrame(orbits, columns=['orbit']).dropna()
        else:
            print('ERROR: Failed to fetch the web page. Status code: {response.status_code}')
            # return the input as is
            return scenes

        orbits['mission']    = orbits['orbit'].apply(lambda name: name[:3])
        orbits['time']       = orbits['orbit'].apply(lambda name: datetime.strptime(name.split('_')[5],      '%Y%m%dT%H%M%S'))
        orbits['time_start'] = orbits['orbit'].apply(lambda name: datetime.strptime(name.split('_')[6][1:],  '%Y%m%dT%H%M%S'))
        orbits['time_end']   = orbits['orbit'].apply(lambda name: datetime.strptime(name[:-4].split('_')[7], '%Y%m%dT%H%M%S'))

        for ind, scene in enumerate(scenes.itertuples()):
            #print ('scene', scene)
            if scene.orbit is not None:
                continue
            # look for the orbit file
            orbit = orbits[(orbits.mission == scene.mission)&\
                           (orbits.time_start <= scene.time_start - orbit_start_offset)&\
                           (orbits.time_end   >= scene.time_end   + orbit_end_offset  )].sort_values('time')
            #print ('orbit', orbit)
            # add the recent orbit
            orbit = orbit.tail(1).orbit.item() if len(orbit) >= 1 else None
            if orbit is None:
                continue
            scenes.at[ind, 'orbit'] = url + orbit
        return scenes

    """
    find . -type f -name '*.tiff' -exec basename {} .tiff \; \
        | awk -F'-' -v OFS='_' '{print toupper($1), "IW_SLC__1SDV", toupper($5), toupper($6), $7, toupper($8), $9"X.SAFE"}' \
        | xargs -I {} mkdir -p {}
    """
    def download(self, basedir, scenes, subswaths, polarization='VV', n_jobs=4, skip_exist=True):
        import pandas as pd
        import numpy as np
        import asf_search
        import fnmatch
        import joblib
        from tqdm.auto import tqdm
        import os
        import re
        from datetime import datetime, timedelta
        import warnings
        # supress asf_search 'UserWarning: File already exists, skipping download'
        warnings.filterwarnings("ignore", category=UserWarning)

        # create the directory if needed
        os.makedirs(basedir, exist_ok=True)

        # collect all the existing files once
        files = os.listdir(basedir)
        #print ('files', len(files))

        # remove all orbits and re-download the same or new ones
        if not skip_exist:
            orbits = [res.group(0) if res else None for file in files for res in [re.search(self.pattern_orbit, file)]]
            orbits = list(filter(None, orbits))
            for orbit in orbits:
                os.remove(os.path.join(basedir, orbit))

        # detect all the local orbits for all scenes
        scenes_orbits = []
        # skip existing scenes
        if skip_exist:
            # output dataframe including scenes and orbits so the both required
            # missed scene or orbit means scene as missed
            scenes_missed = []
            for scene in scenes:
                # check scene
                if not f'{scene}.SAFE' in files:
                    scenes_missed.append(scene)

                # check orbit
                mission = scene[:3]
                dt = datetime.strptime(scene.split('_')[5], '%Y%m%dT%H%M%S')
                day        = dt.strftime('%Y%m%d')
                day_before = (dt + timedelta(days=-1)).strftime('%Y%m%d')
                day_after  = (dt + timedelta(days=1)).strftime('%Y%m%d')
                # check presision orbits
                pattern    = self.template_orbit.format(mission=mission, date_start=day_before, date_end=day_after)
                #print ('Precision orbit pattern', pattern)
                filenames  = sorted(filter(re.compile(pattern).match, files))
                [scenes_orbits.append((scene, filename)) for filename in filenames]
                #print ('filenames', filenames)
                # check approximate orbits when start time is right after midnight and the start date is day before
                if len(filenames) == 0:
                    #print ('Precision orbit not found')
                    pattern   = self.template_orbit.format(mission=mission, date_start=day_before, date_end=day)
                    #print ('Approximate orbit pattern', pattern)
                    filenames = sorted(filter(re.compile(pattern).match, files))
                    [scenes_orbits.append((scene, filename)) for filename in filenames]
                # check approximate orbits
                if len(filenames) == 0:
                    #print ('Precision orbit not found')
                    pattern   = self.template_orbit.format(mission=mission, date_start=day, date_end=day)
                    #print ('Approximate orbit pattern', pattern)
                    filenames = sorted(filter(re.compile(pattern).match, files))
                    [scenes_orbits.append((scene, filename)) for filename in filenames]
                # orbit file is missed when presision and approximate orbits not found
                if len(filenames) == 0:
                    #print ('Approximate orbit not found')
                    scenes_missed.append(scene)

            #print ('scenes_missed', len(scenes_missed))
            scenes = np.unique(scenes_missed)
            #print ('scenes_orbits', scenes_orbits)

        def get_url(scene):
            return self.template_url.format(satellite=scene[2:3], scene=scene)

        def get_patterns(subswaths, polarization, mission):
            #print (f'get_patterns: {subswaths}, {polarization}, {mission}')
            assert len(mission) == 3 and mission[:2]=='S1', \
                f'ERROR: mission name is invalid: {mission}. Expected names like "S1A", "S1B", etc.'
            outs = []
            for subswath in str(subswaths):
                pattern = self.template_safe.format(mission=mission.lower(),
                                                    subswath=subswath,
                                                    polarization=polarization.lower())
                outs.append(pattern)
            return outs

        def download_scene(scene, subswaths, polarization, basedir, session):
            # define scene zip url
            url = get_url(scene)
            #print (f'download_scene: {url}, {patterns}, {basedir}, {session}')
            patterns = get_patterns(subswaths, polarization, mission=scene[:3])
            #print ('patterns', patterns)
            with asf_search.remotezip(url, session) as remotezip:
                filenames = remotezip.namelist()
                #print ('filenames', filenames)
                for pattern in patterns:
                    #print ('pattern', pattern)
                    matching = [filename for filename in filenames if fnmatch.fnmatch(filename, pattern)]
                    for filename in matching:
                        filesize = remotezip.getinfo(filename).file_size
                        #print (filename, filesize)
                        fullname = os.path.join(basedir, filename)
                        if os.path.exists(fullname) and os.path.getsize(fullname) == filesize:
                            #print (f'pass {fullname}')
                            pass
                        else:
                            #print (f'download {fullname}')
                            #with open(fullname, 'wb') as file:
                            #    file.write(remotezip.read(filename))
                            remotezip.extract(filename, basedir)

        # do not initialize internet connection, it allows to work offline when all the scenes and orbits already available
        if len(scenes) == 0:
            return

        # prepare authorized connection
        session = asf_search.ASFSession().auth_with_creds(self.username, self.password)

        # download scenes
        with self.tqdm_joblib(tqdm(desc='ASF Downloading Sentinel-1 SLC', total=len(scenes))) as progress_bar:
            joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(download_scene)(scene, subswaths, polarization, basedir, session) for scene in scenes)
        # debug
        #[download_scene(scene, subswaths, polarization, basedir, session) for scene in scenes]

        # parse processed scenes and convert to dataframe
        #print ('scenes', len(scenes))
        scenes = pd.DataFrame(scenes, columns=['scene'])
        scenes['time_start'] = scenes['scene'].apply(lambda name: datetime.strptime(name.split('_')[5], '%Y%m%dT%H%M%S'))
        scenes['time_end']   = scenes['scene'].apply(lambda name: datetime.strptime(name.split('_')[6], '%Y%m%dT%H%M%S'))
        scenes['mission']    = scenes['scene'].apply(lambda name: name[:3])
        scenes['orbit']      = None

        # download orbit catalogs and look for the required scenes orbits
        if len(scenes[scenes['orbit'].isna()]):
            with tqdm(desc='Downloading POEORB catalog:', total=1) as pbar:
                scenes = self._get_orbits(scenes, session, 'POEORB')
                pbar.update(1)
        if len(scenes[scenes['orbit'].isna()]):
            with tqdm(desc='Downloading RESORB catalog:', total=1) as pbar:
                scenes = self._get_orbits(scenes, session, 'RESORB')
                pbar.update(1)
        if len(scenes[scenes['orbit'].isna()]):
            print ('ERROR: some orbits not found locally and in orbit catalogs')

        def download_orbit(scene, url, basedir, session):
            #print ('download_orbit', scene, url)
            # cleanup one or multiple existing orbits
            # orbits can be outdated and have different names
            # detect all the existing orbits for the scene
            orbits = [scene_orbit[1] for scene_orbit in scenes_orbits if scene_orbit[0]==scene]       
            fullnames = [os.path.join(basedir, orbit) for orbit in orbits]
            #print ('fullnames', fullnames)
            # it can be that scene is missed but the orbit esists when skip_exist=True
            for fullname in fullnames:
                if os.path.exists(fullname):
                    os.remove(fullname)
            # download the same or a new orbit 
            # this routine can use multiple threads but it does not provide a progress indicator
            asf_search.download_urls(urls=[url], path=basedir, session=session)

        # download orbits
        with self.tqdm_joblib(tqdm(desc='ASF Downloading Sentinel-1 Orbits', total=len(scenes.orbit))) as progress_bar:
            joblib.Parallel(n_jobs=n_jobs)(joblib.delayed(download_orbit)(scene, url, basedir, session) \
                                           for (scene, url) in zip(scenes.scene, scenes.orbit))
        # debug
    #    [download_orbit(url, basedir, session) for url in scenes.orbit]

        # return the results in a user-friendly dataframe
        scenes['scene'] = scenes.scene.apply(lambda name: self.template_url.format(satellite=name[2:3], scene=name))
        return scenes[['scene', 'orbit']]



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_trans import Stack_trans
from .tqdm_dask import tqdm_dask

class Stack_trans_inv(Stack_trans):

    def get_trans_inv(self):
        """
        Retrieve the inverse transform data.

        This function opens a NetCDF dataset, which contains data mapping from radar
        coordinates to geographical coordinates (from azimuth-range to latitude-longitude domain).

        Parameters
        ----------
        Returns
        -------
        xarray.Dataset
            An xarray dataset with the transform data.

        Examples
        --------
        Get the inverse transform data:
        get_trans_inv()
        """
        return self.open_cube('trans_inv')

    def compute_trans_inv(self, coarsen, trans='auto', interactive=False):
        """
        Retrieve or calculate the transform data. This transform data is then saved as
            a NetCDF file for future use.

            This function generates data mapping from radar coordinates to geographical coordinates.
            The function uses the direct transform data.

        Parameters
        ----------
        coarsen(jdec, idec) : (int, int) , optional
            The decimation factor in the azimuth and range direction. Default is 2.
        interactive : bool, optional
            If True, the computation will be performed interactively and the result will be returned as a delayed object.
            Default is False.

        Note
        ----
        This function operates on the 'trans' grid using NetCDF chunks (specified by 'netcdf_chunksize') rather than
        larger processing chunks. This approach is effective due to on-the-fly index creation for the NetCDF chunks.

        """
        import dask
        import xarray as xr
        import numpy as np
        import warnings
        warnings.filterwarnings('ignore')

        # convert meters or pixels to radar pixels
        coarsen = self.get_coarsen(coarsen)
        # define maximum search radius, radar pixels
        tolerance = 2 * max(coarsen)

        def trans_inv_block(azis, rngs, tolerance, chunksize):
            from scipy.spatial import cKDTree
            # disable "distributed.utils_perf - WARNING - full garbage collections ..."
            from dask.distributed import utils_perf
            utils_perf.disable_gc_diagnosis()
            import warnings
            warnings.filterwarnings('ignore')

            # required one delta around for nearest interpolation and two for linear
            dazi = np.diff(azis)[0]
            drng = np.diff(rngs)[0]
            azis_min = azis.min() - dazi
            azis_max = azis.max() + dazi
            rngs_min = rngs.min() - drng
            rngs_max = rngs.max() + drng
            del dazi, drng
            #print ('azis_min', azis_min, 'azis_max', azis_max)

            # define valid coordinate blocks 
            block_mask = ((trans_amin<=azis_max)&(trans_amax>=azis_min)&(trans_rmin<=rngs_max)&(trans_rmax>=rngs_min)).values
            block_azi, block_rng = trans_amin.shape
            blocks_ys, blocks_xs = np.meshgrid(range(block_azi), range(block_rng), indexing='ij')
            #assert 0, f'blocks_ys, blocks_xs: {blocks_ys[block_mask]}, {blocks_xs[block_mask]}'
            # extract valid coordinates from the defined blocks
            blocks_trans = []
            blocks_lt = []
            blocks_ll = []
            for block_y, block_x in zip(blocks_ys[block_mask], blocks_xs[block_mask]):
                # coordinates
                block_lt, block_ll = [block.ravel() for block in np.meshgrid(lt_blocks[block_y], ll_blocks[block_x], indexing='ij')]
                # variables
                block_trans = trans.isel(lat=slice(chunksize*block_y,chunksize*(block_y+1)),
                                         lon=slice(chunksize*block_x,chunksize*(block_x+1)))[['azi', 'rng', 'ele']]\
                                   .compute(n_workers=1).to_array().values.reshape(3,-1)
                # select valuable coordinates only
                mask = (block_trans[0,:]>=azis_min)&(block_trans[0,:]<=azis_max)&\
                       (block_trans[1,:]>=rngs_min)&(block_trans[1,:]<=rngs_max)
                # ignore block without valid pixels
                if mask[mask].size > 0:
                    # append valid pixels to accumulators
                    blocks_lt.append(block_lt[mask])
                    blocks_ll.append(block_ll[mask])
                    blocks_trans.append(block_trans[:,mask])
                del block_lt, block_ll, block_trans, mask
            del block_mask, block_azi, block_rng, blocks_ys, blocks_xs

            if len(blocks_lt) == 0:
                # this case is possible when DEM is incomplete, and it is not an error
                return np.nan * np.zeros((3, azis.size, rngs.size), np.float32)

            # TEST
            #return np.nan * np.zeros((3, azis.size, rngs.size), np.float32)

            # valid coordinates
            block_lt = np.concatenate(blocks_lt)
            block_ll = np.concatenate(blocks_ll)
            block_trans = np.concatenate(blocks_trans, axis=1)
            del blocks_lt, blocks_ll, blocks_trans

            # perform index search on radar coordinate grid for the nearest geographic coordinates grid pixel
            grid_azi, grid_rng = np.meshgrid(azis, rngs, indexing='ij')
            tree = cKDTree(np.column_stack([block_trans[0], block_trans[1]]), compact_nodes=False, balanced_tree=False)
            distances, indices = tree.query(np.column_stack([grid_azi.ravel(), grid_rng.ravel()]), k=1, workers=1)
            del grid_azi, grid_rng, tree, cKDTree

            # take the nearest pixels coordinates and elevation
            # the only one index search is required to define all the output variables
            grid_lt = block_lt[indices]
            grid_lt[distances>tolerance] = np.nan
            del block_lt
            grid_ll = block_ll[indices]
            grid_ll[distances>tolerance] = np.nan
            del block_ll
            grid_ele = block_trans[2][indices]
            grid_ele[distances>tolerance] = np.nan
            #print ('distance range', distances.min().round(2), distances.max().round(2))
            #assert distances.max() < 2, f'Unexpectedly large distance between radar and geographic coordinate grid pixels (>=2): {distances.max()}'
            del block_trans, indices, distances

            # pack all the outputs into one 3D array
            return np.asarray([grid_lt, grid_ll, grid_ele]).reshape((3, azis.size, rngs.size))

        if isinstance(trans, str) and trans == 'auto':
            # trans.dat - file generated by llt_grid2rat (r a topo lon lat)"
            trans = self.get_trans()

        # calculate indices on the fly
        trans_blocks = trans[['azi', 'rng']].coarsen(lat=self.netcdf_chunksize, lon=self.netcdf_chunksize, boundary='pad')
        #block_min, block_max = dask.compute(trans_blocks.min(), trans_blocks.max())
        # materialize with progress bar indication
        tqdm_dask(trans_blocks_persist := dask.persist(trans_blocks.min(), trans_blocks.max()), desc='Radar Transform Indexing')
        # only convert structure
        block_min, block_max = dask.compute(trans_blocks_persist)[0]
        trans_amin = block_min.azi
        trans_amax = block_max.azi
        trans_rmin = block_min.rng
        trans_rmax = block_max.rng
        del trans_blocks, block_min, block_max
        #print ('trans_amin', trans_amin)

        # split geographic coordinate grid to equal chunks and rest
        #chunks = trans.azi.data.chunks
        #lt_blocks = np.array_split(trans['lat'].values, np.cumsum(chunks[0])[:-1])
        #ll_blocks = np.array_split(trans['lon'].values, np.cumsum(chunks[1])[:-1])
        lt_blocks = np.array_split(trans['lat'].values, np.arange(0, trans['lat'].size, self.netcdf_chunksize)[1:])
        ll_blocks = np.array_split(trans['lon'].values, np.arange(0, trans['lon'].size, self.netcdf_chunksize)[1:])

        # split radar coordinate grid to equal chunks and rest
        azis, rngs = self.define_trans_grid(coarsen)
        azis_blocks = np.array_split(azis, np.arange(0, azis.size, self.netcdf_chunksize)[1:])
        rngs_blocks = np.array_split(rngs, np.arange(0, rngs.size, self.netcdf_chunksize)[1:])
        #print ('azis_blocks.size', len(azis_blocks), 'rngs_blocks.size', len(rngs_blocks))

        blocks_total = []
        for azis_block in azis_blocks:
            blocks = []
            for rngs_block in rngs_blocks:
                block = dask.array.from_delayed(dask.delayed(trans_inv_block, traverse=False)
                                               (azis_block, rngs_block, tolerance, self.netcdf_chunksize),
                                               shape=(3, azis_block.size, rngs_block.size), dtype=np.float32)
                blocks.append(block)
                del block
            blocks_total.append(blocks)
            del blocks

        trans_inv_dask = dask.array.block(blocks_total)
        del blocks_total
        coords = {'y': azis, 'x': rngs}
        trans_inv = xr.Dataset({key: xr.DataArray(trans_inv_dask[idx],  coords=coords)
                                for idx, key in enumerate(['lt', 'll', 'ele'])})
        del trans_inv_dask

        # calculate indices
        trans_inv['lt_min'] = trans_inv.lt.min('x')
        trans_inv['lt_max'] = trans_inv.lt.max('x')
        trans_inv['ll_min'] = trans_inv.ll.min('y')
        trans_inv['ll_max'] = trans_inv.ll.max('y')

        # add target geographic coordinate grid for the user defined spacing (coarsen)
        trans_inv['lat'] = trans['lat'].values
        trans_inv['lon'] = trans['lon'].values

        if interactive:
            return trans_inv
        # rename to save lazy NetCDF preventing broken coordinates (y,y)
        return self.save_cube(trans_inv, 'trans_inv', f'Radar Inverse Transform Computing')



# ----------------------------------------------------------------------------
# PyGMTSAR
# 
# This file is part of the PyGMTSAR project: https://github.com/mobigroup/gmtsar
# 
# Copyright (c) 2023, Alexey Pechnikov
# 
# Licensed under the BSD 3-Clause License (see LICENSE for details)
# ----------------------------------------------------------------------------
from .Stack_unwrap_snaphu import Stack_unwrap_snaphu
# required for function decorators
from numba import jit
# import directive is not compatible to numba
import numpy as np

class Stack_unwrap(Stack_unwrap_snaphu):

    @staticmethod
    def wrap(data_pairs):
        import xarray as xr
        import numpy as np
        import dask

        if isinstance(data_pairs, xr.DataArray):
            return xr.DataArray(dask.array.mod(data_pairs.data + np.pi, 2 * np.pi) - np.pi, data_pairs.coords)\
                .rename(data_pairs.name)
        return np.mod(data_pairs + np.pi, 2 * np.pi) - np.pi

    # weight is used only to filter NaNs in data
    @staticmethod
    @jit(nopython=True, nogil=True)
    def unwrap_pairs(data, weight, matrix):
        # import directive is not compatible to numba
        #import numpy as np

        # initial data size
        data_size = data.size

        if np.all(np.isnan(data)) or (weight is not None and np.all(np.isnan(weight))):
            return np.nan * np.zeros(data_size)

        # the data variable will be modified and returned as the function output
        data = data.copy()
        if weight is not None:
            nanmask = np.isnan(data) | np.isnan(weight)
        else:
            nanmask = np.isnan(data)
        if np.all(nanmask):
            return np.nan * np.zeros(data_size)
        data = data[~nanmask].copy()
        if weight is not None:
            weight = weight[~nanmask]
        matrix = matrix[~nanmask,:]

        #data = phase_pairs.values
        pair_sum = matrix.sum(axis=1)
        # tidal displacements for SBAS pairs
        #tidal_pairs = (matrix * np.concatenate([[0], np.diff(tidal)])).sum(axis=1)
        # wrapped phase pairs (phase_tidal)
        #data = (np.mod((phase_pairs - tidal_pairs) + np.pi, 2 * np.pi) - np.pi)

        # pairs do not require unwrapping
        pairs_ok = []

        # check all compound pairs vs single pairs: only detect all not wrapped
        for ndate in np.unique(pair_sum)[1:]:
            for pair_idx in np.where(pair_sum==ndate)[0]:
                #print (pair_idx, matrix[pair_idx])
                matching_columns = matrix[pair_idx] == 1
                #print ('matching_columns', matching_columns)
                # works for dates = 2
                #matching_rows = ((matrix[:, matching_columns] == 1).sum(axis=1) == 1)&((matrix[:, ~matching_columns] == 1).sum(axis=1) == 0)
                matching_rows = ((matrix[:, matching_columns] >= 1).sum(axis=1) == 1)&((matrix[:, ~matching_columns] == 1).sum(axis=1) == 0)
                #print ('matching_rows', matching_rows)
                matching_matrix = matrix[:,matching_columns] * matching_rows[:,None]
                #row_indices = np.where(matching_rows)[0]
                #print ('row_indices', row_indices)
                #with np.printoptions(threshold=np.inf, linewidth=np.inf):
                #    print (matching_matrix.T)
                value = data[pair_idx]
                values = (matching_matrix * data[:,None])
                #print ('value', value, '?=', values.sum())    
                jump = int(np.round((values.sum() - value) / (2*np.pi)))
                if jump == 0:
                    pairs_ok.append(pair_idx)
                    pairs_ok.extend(np.where(matching_rows)[0])
                    #print ()
                    continue

                #print ('jump', jump)
                if abs(value) > abs(value + 2*np.pi*jump):
                    #print (f'JUMP {ndate}:', jump)
                    #jump, phase.values[pair_idx] + 2*np.pi*jump
                    #print ('value', value, '=>', value + 2*np.pi*jump, '=', values.sum())
                    pass
                else:
                    #print (f'JUMP singles:', jump)
                    pass
                #print (values[matching_rows])
                # check for wrapping
                valid_values = values[matching_rows].ravel()
                #maxdiff = abs(np.diff(valid_values[valid_values!=0])).max()
                #print ('maxdiff', maxdiff, maxdiff >= np.pi)
                #print ()
                #break
        #print ('pairs_ok', pairs_ok)
        #print ('==================')

        # check all compound pairs vs single pairs: fix wrapped compound using not wrapped singles only
        for ndate in np.unique(pair_sum)[1:]:
            for pair_idx in np.where(pair_sum==ndate)[0]:
                if pair_idx in pairs_ok:
                    continue
                #print (pair_idx, matrix[pair_idx])
                matching_columns = matrix[pair_idx] == 1
                #print ('matching_columns', matching_columns)
                # works for dates = 2
                #matching_rows = ((matrix[:, matching_columns] == 1).sum(axis=1) == 1)&((matrix[:, ~matching_columns] == 1).sum(axis=1) == 0)
                matching_rows = ((matrix[:, matching_columns] >= 1).sum(axis=1) == 1)&((matrix[:, ~matching_columns] == 1).sum(axis=1) == 0)
                #print ('matching_rows', matching_rows)
                matching_matrix = matrix[:,matching_columns] * matching_rows[:,None]
                #row_indices = np.where(matching_rows)[0]
                #print ('row_indices', row_indices)
                #with np.printoptions(threshold=np.inf, linewidth=np.inf):
                #    print (matching_matrix.T)
                #print ('matching_matrix', matching_matrix.shape)

                pairs_single_not_valid = [pair for pair in np.where(matching_rows)[0] if not pair in pairs_ok]
                if len(pairs_single_not_valid) > 0:
                    # some of single-pairs requires unwrapping, miss the compound segment processing
                    #print ('ERROR')
                    #print ()
                    continue

                value = data[pair_idx]
                values = (matching_matrix * data[:,None])
                #print ('value', value, '?=', values.sum())    
                jump = int(np.round((values.sum() - value) / (2*np.pi)))
                #print (f'JUMP {ndate}:', jump)
                data[pair_idx] += 2*np.pi*jump
                pairs_ok.append(pair_idx)
                #print ()
                continue
        #print ('==================')

        # check all compound pairs vs single pairs
        for ndate in np.unique(pair_sum)[1:]:
            for pair_idx in np.where(pair_sum==ndate)[0]:
                if pair_idx in pairs_ok:
                    continue
                #print (pair_idx, matrix[pair_idx])
                matching_columns = matrix[pair_idx] == 1
                #print ('matching_columns', matching_columns)
                # works for dates = 2
                #matching_rows = ((matrix[:, matching_columns] == 1).sum(axis=1) == 1)&((matrix[:, ~matching_columns] == 1).sum(axis=1) == 0)
                matching_rows = ((matrix[:, matching_columns] >= 1).sum(axis=1) == 1)&((matrix[:, ~matching_columns] == 1).sum(axis=1) == 0)
                #print ('matching_rows', matching_rows)
                matching_matrix = matrix[:,matching_columns] * matching_rows[:,None]
                #row_indices = np.where(matching_rows)[0]
                #print ('row_indices', row_indices)
                #with np.printoptions(threshold=np.inf, linewidth=np.inf):
                #    print (matching_matrix.T)
                value = data[pair_idx]
                values = (matching_matrix * data[:,None])
                #print ('value', value, '???=', values.sum())    
                jump = int(np.round((values.sum() - value) / (2*np.pi)))
                #print ('jump', jump)
                if jump != 0:
                    #print (f'JUMP {ndate}:', jump)
                    #jump, phase.values[pair_idx] + 2*np.pi*jump
                    #print ('value', value, '=>', value + 2*np.pi*jump, '=', values.sum())
                    pass
                #print (values[matching_rows])
                # unwrap
                data[pair_idx] += 2*np.pi*jump
                pairs_ok.append(pair_idx)
                pairs_ok.extend(np.where(matching_rows)[0])
                #print ()
                #break

        # validity mask
        mask = [idx in pairs_ok for idx in range(data.size)]
        data = np.where(mask, data, np.nan)
        out = np.nan * np.zeros(data_size)
        out[~nanmask] = data
        return out

    def unwrap1d(self, data, weight=None):
        import xarray as xr
        import numpy as np

        chunks_z, chunks_y, chunks_x = data.chunks
        if np.max(chunks_y) > self.netcdf_chunksize or np.max(chunks_x) > self.netcdf_chunksize:
            print (f'Note: data chunk size ({np.max(chunks_y)}, {np.max(chunks_x)}) is too large for stack processing')
            chunks_y = chunks_x = self.netcdf_chunksize//2
            print (f'Note: auto tune data chunk size to a half of NetCDF chunk: ({chunks_y}, {chunks_x})')
        else:
            # use the existing data chunks size
            chunks_y = np.max(chunks_y)
            chunks_x = np.max(chunks_x)

        pairs = self.get_pairs(data)
        matrix = self.lstsq_matrix(pairs)

        # xarray wrapper
        model = xr.apply_ufunc(
            self.unwrap_pairs,
            data.chunk(dict(pair=-1, y=chunks_y, x=chunks_x)),
            weight.chunk(dict(pair=-1, y=chunks_y, x=chunks_x)) if weight is not None else weight,
            dask='parallelized',
            vectorize=True,
            input_core_dims=[['pair'], ['pair']] if weight is not None else [['pair'], []],
            output_core_dims=[['pair']],
            output_dtypes=[np.float32],
            kwargs={'matrix': matrix}
        ).transpose('pair',...)

        return model.rename('unwrap')

    def unwrap_snaphu(self, phase, weight=None, conf=None, conncomp=False):
        """
        Limit number of processes for tiled multicore SNAPHU configuration:
        with dask.config.set(scheduler='single-threaded'):
            stack.unwrap2d_snaphu()...).phase.compute()
        or
        with dask.config.set(scheduler='threads', num_workers=2):
            stack.unwrap2d_snaphu()...).phase.plot.imshow()
        See for reference: https://docs.dask.org/en/stable/scheduler-overview.html
        """
        import xarray as xr
        import numpy as np
        import dask

        # output dataset variables
        keys = {0: 'phase', 1: 'conncomp'}

        if len(phase.dims) == 2:
            stackvar = None
        else:
            stackvar = phase.dims[0]

        if weight is not None:
            assert phase.shape == weight.shape, 'ERROR: phase and weight variables have different shape'

        def _snaphu(ind):
            ds = self.snaphu(phase.isel({stackvar: ind}) if stackvar is not None else phase,
                             weight.isel({stackvar: ind})  if stackvar is not None and weight is not None else weight,
                             conf=conf, conncomp=conncomp)
            if conncomp:
                # # select the largest connected component
#                 hist = np.unique(ds.conncomp.values, return_counts=True)
#                 idxmax = np.argmax(hist[1])
#                 valmax = hist[0][idxmax]
#                 # select unwrap phase for the largest connected area
#                 return np.stack([ds.phase.where(ds.conncomp==valmax).values, ds.conncomp.values])
                return np.stack([ds.phase.values, ds.conncomp.values])
            return ds.phase.values[None,]

        stack =[]
        for ind in range(len(phase) if stackvar is not None else 1):
            block = dask.array.from_delayed(dask.delayed(_snaphu)(ind),
                        shape=(2 if conncomp else 1, *(phase.shape[1:] if stackvar is not None else phase.shape)),
                        dtype=np.float32)
            stack.append(block)
            del block
        dask_block = dask.array.concatenate(stack)
        del stack
        if stackvar is not None:
            ds = xr.merge([xr.DataArray(dask_block[idx::2 if conncomp else 1], coords=phase.coords).rename(keys[idx])
                               for idx in range(2 if conncomp else 1)])
        else:
            ds = xr.merge([xr.DataArray(block, coords=phase.coords).rename(keys[idx])
                           for idx, block in enumerate(dask_block)])
        del dask_block
        return ds

    def interpolate_nearest(self, data, search_radius_pixels=None):
        """
        Perform nearest neighbor interpolation on each 2D grid in a 3D grid stack.

        Parameters
        ----------
        data : xarray.DataArray
            The input 3D grid stack to be interpolated, with dimensions (pair, y, x).
        search_radius_pixels : int, optional
            The interpolation distance in pixels. If not provided, the default is set to the chunksize of the Stack object.

        Returns
        -------
        xarray.DataArray
            The interpolated 3D grid stack.

        Examples
        --------
        Fill gaps in the specified grid stack using nearest neighbor interpolation:
        stack.interpolate_nearest(intf)

        Notes
        -----
        This method performs nearest neighbor interpolation on each 2D grid (y, x) in a 3D grid stack (pair, y, x). It replaces the NaN values in each 2D grid with the nearest non-NaN values. The interpolation is performed within a specified search radius in pixels for each grid. If a search radius is not provided, the default search radius is set to the chunksize of the Stack object.
        """
        import xarray as xr

        assert data.dims == ('pair', 'y', 'x'), 'Input data must have dimensions (pair, y, x)'

        interpolated = [self.nearest_grid(data.sel(pair=pair), search_radius_pixels) for pair in data.pair]
        return xr.concat(interpolated, dim='pair')

    @staticmethod
    def conncomp_main(data):
        import xarray as xr
        import numpy as np

        # Function to find the mode (most frequent value)
        def find_mode(array):
            values, counts = np.unique(array[~np.isnan(array)], return_counts=True)
            max_count_index = np.argmax(counts)
            return values[max_count_index] if counts.size > 0 else np.nan

        # Apply the function along the 'pair' dimension
        maincomps =  xr.apply_ufunc(find_mode, data.conncomp, input_core_dims=[['y', 'x']],
                                    vectorize=True, dask='parallelized', output_dtypes=[float])
        return data.where(data.conncomp==maincomps)

    def plot_conncomps(self, data, caption='Connected Components', cols=4, size=4, nbins=5, aspect=1.2, y=1.05,
                       vmin=0, vmax=10, cmap='tab10_r'):
        import matplotlib.pyplot as plt

        # multi-plots ineffective for linked lazy data
        fg = data.plot.imshow(
            col='pair',
            col_wrap=cols, size=size, aspect=aspect,
            cmap=cmap, vmin=0, vmax=10
        )
        fg.set_axis_labels('Range', 'Azimuth')
        fg.set_ticks(max_xticks=nbins, max_yticks=nbins)
        fg.fig.suptitle(caption, y=y)
